#
# Copyright (c) 2014 VMware, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
test_congress
----------------------------------

Tests for `congress` module.
"""
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import mock
import neutronclient.v2_0
from oslo_log import log as logging

from congress.api import base as api_base
from congress.common import config
from congress.datalog import compile
from congress.datasources import neutronv2_driver
from congress.datasources import nova_driver
from congress.db import db_library_policies
from congress.tests.api import base as tests_api_base
from congress.tests import base
from congress.tests.datasources import fakes
from congress.tests import helper


LOG = logging.getLogger(__name__)


class BaseTestPolicyCongress(base.SqlTestCase):

    def setUp(self):
        super(BaseTestPolicyCongress, self).setUp()
        self.services = tests_api_base.setup_config(with_fake_datasource=False)
        self.api = self.services['api']
        self.node = self.services['node']
        self.engine = self.services['engine']
        self.library = self.services['library']

        self.neutronv2 = self._create_neutron_mock('neutron')

    def tearDown(self):
        self.node.stop()
        super(BaseTestPolicyCongress, self).tearDown()

    def _create_neutron_mock(self, name):
        # Register Neutron service
        args = helper.datasource_openstack_args()
        neutronv2 = neutronv2_driver.NeutronV2Driver(name, args=args)
        # FIXME(ekcs): this is a hack to prevent the synchronizer from
        # attempting to delete this DSD because it's not in DB
        neutronv2.type = 'no_sync_datasource_driver'
        neutron_mock = mock.MagicMock(spec=neutronclient.v2_0.client.Client)
        neutronv2.neutron = neutron_mock

        # initialize neutron_mocks
        network1 = fakes.network_response
        port_response = fakes.port_response
        router_response = fakes.router_response
        sg_group_response = fakes.security_group_response
        neutron_mock.list_networks.return_value = network1
        neutron_mock.list_ports.return_value = port_response
        neutron_mock.list_routers.return_value = router_response
        neutron_mock.list_security_groups.return_value = sg_group_response
        self.node.register_service(neutronv2)
        return neutronv2


class TestCongress(BaseTestPolicyCongress):

    def setUp(self):
        """Setup tests that use multiple mock neutron instances."""
        super(TestCongress, self).setUp()

        # clear the library policies loaded on startup
        db_library_policies.delete_policies()

    def tearDown(self):
        super(TestCongress, self).tearDown()

    def setup_config(self):
        args = ['--config-file', helper.etcdir('congress.conf.test')]
        config.init(args)

    def test_startup(self):
        self.assertIsNotNone(self.services['api'])
        self.assertIsNotNone(self.services['engine'])
        self.assertIsNotNone(self.services['library'])
        self.assertIsNotNone(self.services['engine'].node)

    def test_policy(self):
        self.create_policy('alpha')
        self.insert_rule('q(1, 2) :- true', 'alpha')
        self.insert_rule('q(2, 3) :- true', 'alpha')
        helper.retry_check_function_return_value(
            lambda: sorted(self.query('q', 'alpha')['results'],
                           key=lambda x: x['data']),
            sorted([{'data': (1, 2)}, {'data': (2, 3)}],
                   key=lambda x: x['data']))
        helper.retry_check_function_return_value(
            lambda: list(self.query('q', 'alpha').keys()),
            ['results'])

    def test_policy_datasource(self):
        self.create_policy('alpha')
        self.create_fake_datasource('fake')
        data = self.node.service_object('fake')
        data.state = {'fake_table': set([(1, 2)])}

        data.poll()
        self.insert_rule('q(x) :- fake:fake_table(x,y)', 'alpha')
        helper.retry_check_function_return_value(
            lambda: self.query('q', 'alpha'), {'results': [{'data': (1,)}]})

        # TODO(dse2): enable rules to be inserted before data created.
        #  Maybe just have subscription handle errors graceful when
        #  asking for a snapshot and return [].
        # self.insert_rule('p(x) :- fake:fake_table(x)', 'alpha')

    def test_policy_create_from_library(self):
        def adjust_for_comparison(rules):
            # compile rule string into rule object
            # replace dict with tuple for sorting
            # 'id' field implicitly dropped if present
            rules = [(compile.parse1(rule['rule']), rule['name'],
                      rule['comment']) for rule in rules]

            # sort lists for comparison
            return sorted(rules)

        test_policy = {
            "name": "test_policy",
            "description": "test policy description",
            "kind": "nonrecursive",
            "abbreviation": "abbr",
            "rules": [{"rule": "p(x) :- q(x)", "comment": "test comment",
                       "name": "test name"},
                      {"rule": "p(x) :- q2(x)", "comment": "test comment2",
                       "name": "test name2"}]
        }
        test_policy_id, test_policy_obj = self.api[
            'api-library-policy'].add_item(test_policy, {})

        add_policy_id, add_policy_obj = self.api['api-policy'].add_item(
            None, {'library_policy': test_policy_id})

        test_policy['id'] = add_policy_id

        # adjust for comparison
        test_policy['owner_id'] = 'user'
        test_policy['rules'] = adjust_for_comparison(test_policy['rules'])

        add_policy_obj['rules'] = adjust_for_comparison(
            add_policy_obj['rules'])

        self.assertEqual(add_policy_obj, test_policy)

        context = {'policy_id': test_policy['name']}
        rules = self.api['api-rule'].get_items({}, context)['results']
        rules = adjust_for_comparison(rules)
        self.assertEqual(rules, test_policy['rules'])

        res = self.api['api-policy'].get_items({})['results']
        del test_policy['rules']
        self.assertIn(test_policy, res)

    def test_policy_create_with_rules(self):
        def adjust_for_comparison(rules):
            # compile rule string into rule object
            # replace dict with tuple for sorting
            # 'id' field implicitly dropped if present
            rules = [(compile.parse1(rule['rule']), rule['name'],
                      rule['comment']) for rule in rules]

            # sort lists for comparison
            return sorted(rules)

        test_policy = {
            "name": "test_policy",
            "description": "test policy description",
            "kind": "nonrecursive",
            "abbreviation": "abbr",
            "rules": [{"rule": "p(x) :- q(x)", "comment": "test comment",
                       "name": "test name"},
                      {"rule": "p(x) :- q2(x)", "comment": "test comment2",
                       "name": "test name2"}]
        }

        add_policy_id, add_policy_obj = self.api['api-policy'].add_item(
            test_policy, {})

        test_policy['id'] = add_policy_id

        # adjust for comparison
        test_policy['owner_id'] = 'user'
        test_policy['rules'] = adjust_for_comparison(test_policy['rules'])

        add_policy_obj['rules'] = adjust_for_comparison(
            add_policy_obj['rules'])

        self.assertEqual(add_policy_obj, test_policy)

        context = {'policy_id': test_policy['name']}
        rules = self.api['api-rule'].get_items({}, context)['results']
        rules = adjust_for_comparison(rules)
        self.assertEqual(rules, test_policy['rules'])

        res = self.api['api-policy'].get_items({})['results']
        del test_policy['rules']
        self.assertIn(test_policy, res)

    def create_policy(self, name):
        return self.api['api-policy'].add_item({'name': name}, {})

    def create_policy_from_obj(self, policy_obj):
        return self.api['api-policy'].add_item(policy_obj, {})

    def insert_rule(self, rule, policy):
        context = {'policy_id': policy}
        return self.api['api-rule'].add_item(
            {'rule': rule}, {}, context=context)

    def create_fake_datasource(self, name):
        item = {'name': name,
                'driver': 'fake_datasource',
                'description': 'hello world!',
                'enabled': True,
                'type': None,
                'config': {'auth_url': 'foo',
                           'username': 'armax',
                           'password': '<hidden>',
                           'tenant_name': 'armax'}}

        return self.api['api-datasource'].add_item(item, params={})

    def query(self, tablename, policyname):
        context = {'policy_id': policyname,
                   'table_id': tablename}
        return self.api['api-row'].get_items({}, context)

    def test_rule_insert_delete(self):
        self.api['api-policy'].add_item({'name': 'alice'}, {})
        context = {'policy_id': 'alice'}
        (id1, _) = self.api['api-rule'].add_item(
            {'rule': 'p(x) :- plus(y, 1, x), q(y)'}, {}, context=context)
        ds = self.api['api-rule'].get_items({}, context)['results']
        self.assertEqual(len(ds), 1)
        self.api['api-rule'].delete_item(id1, {}, context)
        ds = self.engine.policy_object('alice').content()
        self.assertEqual(len(ds), 0)

    def test_datasource_request_refresh(self):
        # neutron polls automatically here, which is why register_service
        # starts its service.
        neutron = self.neutronv2
        neutron.stop()

        self.assertEqual(neutron.refresh_request_queue.qsize(), 0)
        neutron.request_refresh()
        self.assertEqual(neutron.refresh_request_queue.qsize(), 1)
        neutron.start()

        neutron.request_refresh()
        f = lambda: neutron.refresh_request_queue.qsize()
        helper.retry_check_function_return_value(f, 0)

    def test_datasource_poll(self):
        neutron = self.neutronv2
        neutron.stop()
        neutron._translate_ports({'ports': []})
        self.assertEqual(len(neutron.state['ports']), 0)
        neutron.start()
        f = lambda: len(neutron.state['ports'])
        helper.retry_check_function_return_value_not_eq(f, 0)

    def test_library_service(self):
        # NOTE(ekcs): only the most basic test right now, more detailed testing
        # done in test_library_service.py
        res = self.library.get_policies()
        self.assertEqual(res, [])

    def test_load_library_policies(self):
        self.assertEqual(self.library.load_policies_from_files(), 0)


class APILocalRouting(BaseTestPolicyCongress):

    def setUp(self):
        super(APILocalRouting, self).setUp()

        # set up second API+PE node
        self.services = tests_api_base.setup_config(
            with_fake_datasource=False, node_id='testnode2',
            same_partition_as_node=self.node)
        self.api2 = self.services['api']
        self.node2 = self.services['node']
        self.engine2 = self.services['engine']
        self.data = self.services['data']

        # add different data to two PE instances
        # going directly to agnostic not via API to make sure faulty API
        # routing (subject of the test) would not affect test accuracy
        self.engine.create_policy('policy')
        self.engine2.create_policy('policy')
        self.engine.insert('p(1) :- NOT q()', 'policy')
        # self.engine1.insert('p(1)', 'policy')
        self.engine2.insert('p(2) :- NOT q()', 'policy')
        self.engine2.insert('p(3) :- NOT q()', 'policy')

    def test_intranode_pe_routing(self):
        for i in range(0, 5):  # run multiple times (non-determinism)
            result = self.api['api-row'].get_items(
                {}, {'policy_id': 'policy', 'table_id': 'p'})
            self.assertEqual(len(result['results']), 1)
            result = self.api2['api-row'].get_items(
                {}, {'policy_id': 'policy', 'table_id': 'p'})
            self.assertEqual(len(result['results']), 2)

    def test_non_PE_service_reachable(self):
        # intranode
        result = self.api['api-row'].get_items(
            {}, {'ds_id': 'neutron', 'table_id': 'ports'})
        self.assertEqual(len(result['results']), 1)

        # internode
        result = self.api2['api-row'].get_items(
            {}, {'ds_id': 'neutron', 'table_id': 'ports'})
        self.assertEqual(len(result['results']), 1)

    def test_internode_pe_routing(self):
        '''test reach internode PE when intranode PE not available'''
        self.node.unregister_service(api_base.ENGINE_SERVICE_ID)
        result = self.api['api-row'].get_items(
            {}, {'policy_id': 'policy', 'table_id': 'p'})
        self.assertEqual(len(result['results']), 2)
        result = self.api2['api-row'].get_items(
            {}, {'policy_id': 'policy', 'table_id': 'p'})
        self.assertEqual(len(result['results']), 2)


class TestPolicyExecute(BaseTestPolicyCongress):

    def setUp(self):
        super(TestPolicyExecute, self).setUp()
        self.nova = self._register_test_datasource('nova')

    def _register_test_datasource(self, name):
        args = helper.datasource_openstack_args()
        if name == 'nova':
            ds = nova_driver.NovaDriver('nova', args=args)
        if name == 'neutron':
            ds = neutronv2_driver.NeutronV2Driver('neutron', args=args)
        ds.update_from_datasource = mock.MagicMock()
        return ds

    def test_policy_execute(self):
        class NovaClient(object):
            def __init__(self, testkey):
                self.testkey = testkey

            def disconnectNetwork(self, arg1):
                LOG.info("disconnectNetwork called on %s", arg1)
                self.testkey = "arg1=%s" % arg1

        nova_client = NovaClient("testing")
        nova = self.nova
        nova.nova_client = nova_client
        self.node.register_service(nova)

        # insert rule and data
        self.api['api-policy'].add_item({'name': 'alice'}, {})
        (id1, _) = self.api['api-rule'].add_item(
            {'rule': 'execute[nova:disconnectNetwork(x)] :- q(x)'}, {},
            context={'policy_id': 'alice'})
        self.assertEqual(len(self.engine.logger.messages), 0)
        (id2, _) = self.api['api-rule'].add_item(
            {'rule': 'q(1)'}, {}, context={'policy_id': 'alice'})
        self.assertEqual(len(self.engine.logger.messages), 1)
        ans = "arg1=1"
        f = lambda: nova.nova_client.testkey
        helper.retry_check_function_return_value(f, ans)

        # insert more data
        self.api['api-rule'].add_item(
            {'rule': 'q(2)'}, {}, context={'policy_id': 'alice'})
        self.assertEqual(len(self.engine.logger.messages), 2)
        ans = "arg1=2"
        f = lambda: nova.nova_client.testkey
        helper.retry_check_function_return_value(f, ans)

        # insert irrelevant data
        self.api['api-rule'].add_item(
            {'rule': 'r(3)'}, {}, context={'policy_id': 'alice'})
        self.assertEqual(len(self.engine.logger.messages), 2)

        # delete relevant data
        self.api['api-rule'].delete_item(
            id2, {}, context={'policy_id': 'alice'})
        self.assertEqual(len(self.engine.logger.messages), 2)

        # delete policy rule
        self.api['api-rule'].delete_item(
            id1, {}, context={'policy_id': 'alice'})
        self.assertEqual(len(self.engine.logger.messages), 2)

    def test_policy_execute_data_first(self):
        class NovaClient(object):
            def __init__(self, testkey):
                self.testkey = testkey

            def disconnectNetwork(self, arg1):
                LOG.info("disconnectNetwork called on %s", arg1)
                self.testkey = "arg1=%s" % arg1

        nova_client = NovaClient(None)
        nova = self.nova
        nova.nova_client = nova_client
        self.node.register_service(nova)

        # insert rule and data
        self.api['api-policy'].add_item({'name': 'alice'}, {})
        self.api['api-rule'].add_item(
            {'rule': 'q(1)'}, {}, context={'policy_id': 'alice'})
        self.assertEqual(len(self.engine.logger.messages), 0)
        self.api['api-rule'].add_item(
            {'rule': 'execute[nova:disconnectNetwork(x)] :- q(x)'}, {},
            context={'policy_id': 'alice'})
        self.assertEqual(len(self.engine.logger.messages), 1)
        ans = "arg1=1"
        f = lambda: nova.nova_client.testkey
        helper.retry_check_function_return_value(f, ans)

    def test_policy_execute_dotted(self):
        class NovaClient(object):
            def __init__(self, testkey):
                self.testkey = testkey
                self.servers = ServersClass()

        class ServersClass(object):
            def __init__(self):
                self.ServerManager = ServerManagerClass()

        class ServerManagerClass(object):
            def __init__(self):
                self.testkey = None

            def pause(self, id_):
                self.testkey = "arg1=%s" % id_

        nova_client = NovaClient(None)
        nova = self.nova
        nova.nova_client = nova_client
        self.node.register_service(nova)

        self.api['api-policy'].add_item({'name': 'alice'}, {})
        self.api['api-rule'].add_item(
            {'rule': 'execute[nova:servers.ServerManager.pause(x)] :- q(x)'},
            {}, context={'policy_id': 'alice'})
        self.assertEqual(len(self.engine.logger.messages), 0)
        self.api['api-rule'].add_item(
            {'rule': 'q(1)'}, {}, context={'policy_id': 'alice'})
        self.assertEqual(len(self.engine.logger.messages), 1)
        ans = "arg1=1"
        f = lambda: nova.nova_client.servers.ServerManager.testkey
        helper.retry_check_function_return_value(f, ans)

    def test_policy_execute_no_args(self):
        class NovaClient(object):
            def __init__(self, testkey):
                self.testkey = testkey

            def disconnectNetwork(self):
                LOG.info("disconnectNetwork called")
                self.testkey = "noargs"

        nova_client = NovaClient(None)
        nova = self.nova
        nova.nova_client = nova_client
        self.node.register_service(nova)

        # Note: this probably isn't the behavior we really want.
        #  But at least we have a test documenting that behavior.

        # insert rule and data
        self.api['api-policy'].add_item({'name': 'alice'}, {})
        (id1, rule1) = self.api['api-rule'].add_item(
            {'rule': 'execute[nova:disconnectNetwork()] :- q(x)'}, {},
            context={'policy_id': 'alice'})
        self.assertEqual(len(self.engine.logger.messages), 0)
        (id2, rule2) = self.api['api-rule'].add_item(
            {'rule': 'q(1)'}, {}, context={'policy_id': 'alice'})
        self.assertEqual(len(self.engine.logger.messages), 1)
        ans = "noargs"
        f = lambda: nova.nova_client.testkey
        helper.retry_check_function_return_value(f, ans)

        # insert more data (which DOES NOT cause an execution)
        (id3, rule3) = self.api['api-rule'].add_item(
            {'rule': 'q(2)'}, {}, context={'policy_id': 'alice'})
        self.assertEqual(len(self.engine.logger.messages), 1)

        # delete all data
        self.api['api-rule'].delete_item(
            id2, {}, context={'policy_id': 'alice'})
        self.assertEqual(len(self.engine.logger.messages), 1)

        self.api['api-rule'].delete_item(
            id3, {}, context={'policy_id': 'alice'})
        self.assertEqual(len(self.engine.logger.messages), 1)

        # insert data (which now DOES cause an execution)
        (id4, rule3) = self.api['api-rule'].add_item(
            {'rule': 'q(3)'}, {}, context={'policy_id': 'alice'})
        self.assertEqual(len(self.engine.logger.messages), 2)
        ans = "noargs"
        f = lambda: nova.nova_client.testkey
        helper.retry_check_function_return_value(f, ans)

        # delete policy rule
        self.api['api-rule'].delete_item(
            id1, {}, context={'policy_id': 'alice'})
        self.assertEqual(len(self.engine.logger.messages), 2)

    def test_neutron_policy_execute(self):
        class NeutronClient(object):
            def __init__(self, testkey):
                self.testkey = testkey

            def disconnectNetwork(self, arg1):
                LOG.info("disconnectNetwork called on %s", arg1)
                self.testkey = "arg1=%s" % arg1

        neutron_client = NeutronClient(None)
        neutron = self.neutronv2
        neutron.neutron = neutron_client

        # insert rule and data
        self.api['api-policy'].add_item({'name': 'alice'}, {})
        (id1, _) = self.api['api-rule'].add_item(
            {'rule': 'execute[neutron:disconnectNetwork(x)] :- q(x)'}, {},
            context={'policy_id': 'alice'})
        self.assertEqual(len(self.engine.logger.messages), 0)
        (id2, _) = self.api['api-rule'].add_item(
            {'rule': 'q(1)'}, {}, context={'policy_id': 'alice'})
        self.assertEqual(len(self.engine.logger.messages), 1)
        ans = "arg1=1"
        f = lambda: neutron.neutron.testkey
        helper.retry_check_function_return_value(f, ans)

    def test_neutron_policy_poll_and_subscriptions(self):
        """Test polling and publishing of neutron updates."""
        policy = self.engine.DEFAULT_THEORY
        neutron2 = self._create_neutron_mock('neutron2')
        self.engine.initialize_datasource('neutron',
                                          self.neutronv2.get_schema())
        self.engine.initialize_datasource('neutron2',
                                          self.neutronv2.get_schema())
        str_rule = ('p(x0, y0) :- neutron:networks(x0, x1, x2, x3, x4, x5), '
                    'neutron2:networks(y0, y1, y2, y3, y4, y5)')
        rule = {'rule': str_rule, 'name': 'testrule1', 'comment': 'test'}
        self.api['api-rule'].add_item(rule, {}, context={'policy_id': policy})
        # Test policy subscriptions
        subscriptions = self.engine.subscription_list()
        self.assertEqual(sorted([('neutron', 'networks'),
                         ('neutron2', 'networks')]), sorted(subscriptions))
        # Test multiple instances
        self.neutronv2.poll()
        neutron2.poll()
        ans = ('p("240ff9df-df35-43ae-9df5-27fae87f2492", '
               '  "240ff9df-df35-43ae-9df5-27fae87f2492") ')
        helper.retry_check_db_equal(self.engine, 'p(x, y)', ans, target=policy)



# Copyright (c) 2014 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import mock

# Sample responses from neutron-client, after parsing
network_response = {
    'networks':
        [{'status': 'ACTIVE',
          'subnets': ['4cef03d0-1d02-40bb-8c99-2f442aac6ab0'],
          'name': 'test-network',
          'provider:physical_network': None,
          'admin_state_up': True,
          'tenant_id': '570fe78a1dc54cffa053bd802984ede2',
          'provider:network_type': 'gre',
          'router:external': False,
          'shared': False,
          'id': '240ff9df-df35-43ae-9df5-27fae87f2492',
          'provider:segmentation_id': 4}]}

port_response = {
    "ports":
        [{"status": "ACTIVE",
          "binding:host_id": "havana",
          "name": "",
          "allowed_address_pairs": [],
          "admin_state_up": True,
          "network_id": "240ff9df-df35-43ae-9df5-27fae87f2492",
          "tenant_id": "570fe78a1dc54cffa053bd802984ede2",
          "extra_dhcp_opts": [],
          "binding:vif_type": "ovs",
          "device_owner": "network:router_interface",
          "binding:capabilities": {"port_filter": True},
          "mac_address": "fa:16:3e:ab:90:df",
          "fixed_ips": [
              {"subnet_id": "4cef03d0-1d02-40bb-8c99-2f442aac6ab0",
               "ip_address": "90.0.0.1"},
              {"subnet_id": "5cef03d0-1d02-40bb-8c99-2f442aac6ab0",
               "ip_address": "100.0.0.1"}],
          "id": "0a2ce569-85a8-45ec-abb3-0d4b34ff69ba",
          "security_groups": ['15ea0516-11ec-46e9-9e8e-7d1b6e3d7523',
                              '25ea0516-11ec-46e9-9e8e-7d1b6e3d7523'],
          "device_id": "864e4acf-bf8e-4664-8cf7-ad5daa95681e"}]}


router_response = {
    'routers':
        [{u'status': u'ACTIVE',
          u'external_gateway_info':
            {u'network_id': u'a821b8d3-af1f-4d79-9b8e-3da9674338ae',
             u'enable_snat': True},
          u'name': u'router1',
          u'admin_state_up': True,
          u'tenant_id': u'abb53cc6636848218f46d01f22bf1060',
          u'routes': [],
          u'id': u'4598c424-d608-4366-9beb-139adbd7cff5'}]}

security_group_response = {
    'security_groups':
        [{u'tenant_id': u'abb53cc6636848218f46d01f22bf1060',
          u'name': u'default',
          u'description': u'default',
          u'security_group_rules': [
              {u'remote_group_id': u'9f3860a5-87b1-499c-bf93-5ca3ef247517',
               u'direction': u'ingress',
               u'remote_ip_prefix': None,
               u'protocol': None,
               u'tenant_id': u'abb53cc6636848218f46d01f22bf1060',
               u'port_range_max': None,
               u'security_group_id': u'9f3860a5-87b1-499c-bf93-5ca3ef247517',
               u'port_range_min': None,
               u'ethertype': u'IPv6',
               u'id': u'15ea0516-11ec-46e9-9e8e-7d1b6e3d7523'}],
          u'id': u'9f3860a5-87b1-499c-bf93-5ca3ef247517'}]}


class NovaFakeClient(mock.MagicMock):
    # TODO(rajdeepd): Replace Fake with mocks directly in test_neutron_driver
    def __init__(self, *args, **kwargs):
        super(NovaFakeClient, self).__init__(*args, **kwargs)
        self.servers = mock.MagicMock()
        self.servers.list.return_value = self.get_server_list()
        self.flavors = mock.MagicMock()
        self.flavors.list.return_value = self.get_flavor_list()

        # self.hosts = mock.MagicMock()
        # self.hosts.list.return_value = self.get_host_list()
        self.hypervisors = mock.MagicMock()
        self.hypervisors.list.return_value = self.get_hypervisor_list()
        self.services = mock.MagicMock()
        self.services.list.return_value = self.get_service_list()

        self.availability_zones = mock.MagicMock()
        self.availability_zones.list.return_value = self.get_zone_list()

    def get_mock_server(self, id, name, host_id, status, tenant_id, user_id,
                        flavor, image, zone=None, host_name=None, tags=None):
        server = mock.MagicMock()
        server.id = id
        server.hostId = host_id
        server.tenant_id = tenant_id
        server.user_id = user_id
        server.status = status
        server.name = name
        server.image = image
        server.flavor = flavor
        server.tags = tags if tags else []
        if zone is not None:
            setattr(server, 'OS-EXT-AZ:availability_zone', zone)
        else:
            # This ensures that the magic mock raises an AttributeError
            delattr(server, 'OS-EXT-AZ:availability_zone')
        if host_name is not None:
            setattr(server, 'OS-EXT-SRV-ATTR:hypervisor_hostname',
                    host_name)
        else:
            # This ensures that the magic mock raises an AttributeError
            delattr(server, 'OS-EXT-SRV-ATTR:hypervisor_hostname')
        return server

    def get_server_list(self):
        server_one = (
            self.get_mock_server('1234', 'sample-server',
                                 "e4d909c290d0fb1ca068ffaddf22cbd0",
                                 'BUILD',
                                 '50e14867-7c64-4ec9-be8d-ed2470ca1d24',
                                 '33ea0494-2bdf-4382-a445-9068997430b9',
                                 {"id": "1"}, {"id": "2"}, 'default', 'host1'))

        server_two = (
            self.get_mock_server('5678', 'sample-server2',
                                 "9e107d9d372bb6826bd81d3542a419d6",
                                 'ACTIVE',
                                 '50e14867-7c64-4ec9-be8d-ed2470ca1d24',
                                 '33ea0494-2bdf-4382-a445-9068997430b9',
                                 {"id": "1"}, {"id": "2"},
                                 tags=['tag1', 'tag2']))

        server_three = (
            self.get_mock_server('9012', 'sample-server3',
                                 "9e107d9d372bb6826bd81d3542a419d6",
                                 'ACTIVE',
                                 '50e14867-7c64-4ec9-be8d-ed2470ca1d24',
                                 '33ea0494-2bdf-4382-a445-9068997430b9',
                                 {"id": "1"}, {"id": "2"}, 'foo', 'host2',
                                 tags=['tag1', 'tag2', 'tag3']))

        return [server_one, server_two, server_three]

    def get_flavor(self, id, name, vcpus, ram, disk, ephemeral, rxtx_factor):
        f = mock.MagicMock()
        f.id = id
        f.name = name
        f.vcpus = vcpus
        f.ram = ram
        f.disk = disk
        f.ephemeral = ephemeral
        f.rxtx_factor = rxtx_factor
        return f

    def get_flavor_list(self):
        flavor_one = self.get_flavor("1", "256 MB Server", 1, 256, 10, 10, 1.0)
        flavor_two = self.get_flavor("2", "512 MB Server", 2, 512, 20, 20, 1.0)
        flavor_three = self.get_flavor("3", "128 MB Server", 4, 128, 0, 0, 3.0)
        flavor_four = self.get_flavor("4", "1024 MB Server", 3, 1024, 10, 10,
                                      2.0)

        return [flavor_one, flavor_two, flavor_three, flavor_four]

#    def get_host(self, host_name, service, zone):
#        h = mock.MagicMock()
#        h.host_name = host_name
#        h.service = service
#        h.zone = zone
#        return h
#
#    def get_host_list(self):
#        h_one = self.get_host('host1', 'nova-compute', 'nova1')
#        h_two = self.get_host('host2', 'nova-cert', 'nova1')
#
#        return [h_one, h_two]

    def get_hypervisor(self, host_name, id_, state, status):
        h = mock.MagicMock()
        h.hypervisor_hostname = host_name
        h.id = id_
        h.state = state
        h.status = status
        return h

    def get_hypervisor_list(self, nova_api_version='2.26'):
        from distutils.version import StrictVersion
        if StrictVersion(nova_api_version) <= StrictVersion('2.52'):
            h_one = self.get_hypervisor('host1', 2, 'up', 'enabled')
            h_two = self.get_hypervisor('host2', 3, 'down', 'enabled')
        else:
            h_one = self.get_hypervisor('host1', '2', 'up', 'enabled')
            h_two = self.get_hypervisor('host2', '3', 'down', 'enabled')

        return [h_one, h_two]

    def get_service(self, id, binary, host, zone, status, state,
                    updated_at, disabled_reason):
        s = mock.MagicMock()
        s.id = id
        s.binary = binary
        s.host = host
        s.zone = zone
        s.status = status
        s.state = state
        s.updated_at = updated_at
        s.disabled_reason = disabled_reason

        return s

    def get_service_list(self):
        service_one = self.get_service(1, 'nova-compute', 'nova',
                                       'nova1', 'enabled', 'up',
                                       '2015-07-28T08:28:37.000000', None)
        service_two = self.get_service(2, 'nova-schedule', 'nova',
                                       'nova1', 'disabled', 'up',
                                       '2015-07-28T08:28:38.000000',
                                       'daily maintenance')

        return [service_one, service_two]

    def get_availability_zone(self, name, state):
        zone = mock.MagicMock()
        zone.zoneName = name
        zone.zoneState = state
        return zone

    def get_zone_list(self):
        zone_one = self.get_availability_zone('AZ1', 'available')
        zone_two = self.get_availability_zone('AZ2', 'not available')

        return [zone_one, zone_two]



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# Copyright 2011 Justin Santa Barbara
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Utilities and helper functions."""
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import contextlib
import json
import os
import shutil
import tempfile

from oslo_config import cfg
from oslo_log import log as logging
import six

LOG = logging.getLogger(__name__)

utils_opts = [
    cfg.StrOpt('tempdir',
               help='Explicitly specify the temporary working directory'),
]
CONF = cfg.CONF
CONF.register_opts(utils_opts)


# Note(thread-safety): blocking function
@contextlib.contextmanager
def tempdir(**kwargs):
    argdict = kwargs.copy()
    if 'dir' not in argdict:
        argdict['dir'] = CONF.tempdir
    tmpdir = tempfile.mkdtemp(**argdict)
    try:
        yield tmpdir
    finally:
        try:
            shutil.rmtree(tmpdir)
        except OSError as e:
            LOG.error(('Could not remove tmpdir: %s'), e)


def value_to_congress(value):
    if isinstance(value, six.string_types):
        # TODO(ayip): This throws away high unicode data because congress does
        # not have full support for unicode yet.  We'll need to fix this to
        # handle unicode coming from datasources.
        try:
            six.text_type(value).encode('ascii')
        except UnicodeEncodeError:
            LOG.warning('Ignoring non-ascii characters')
        # Py3: decode back into str for compat (bytes != str)
        return six.text_type(value).encode('ascii', 'ignore').decode('ascii')
    # Check for bool before int, because True and False are also ints.
    elif isinstance(value, bool):
        return str(value)
    elif (isinstance(value, six.integer_types) or
          isinstance(value, float)):
        return value
    return str(value)


def tuple_to_congress(value_tuple):
    return tuple(value_to_congress(v) for v in value_tuple)


# Note(thread-safety): blocking function
def create_datasource_policy(bus, datasource, engine):
    # Get the schema for the datasource using
    # Note(thread-safety): blocking call
    schema = bus.rpc(datasource, 'get_datasource_schema',
                     {'source_id': datasource})
    # Create policy and sets the schema once datasource is created.
    args = {'name': datasource, 'schema': schema}
    # Note(thread-safety): blocking call
    bus.rpc(engine, 'initialize_datasource', args)


def get_root_path():
    return os.path.dirname(os.path.dirname(__file__))


class Location (object):
    """A location in the program source code."""

    __slots__ = ['line', 'col']

    def __init__(self, line=None, col=None, obj=None):
        try:
            self.line = obj.location.line
            self.col = obj.location.col
        except AttributeError:
            pass
        self.col = col
        self.line = line

    def __str__(self):
        s = ""
        if self.line is not None:
            s += " line: {}".format(self.line)
        if self.col is not None:
            s += " col: {}".format(self.col)
        return s

    def __repr__(self):
        return "Location(line={}, col={})".format(
            repr(self.line), repr(self.col))

    def __hash__(self):
        return hash(('Location', hash(self.line), hash(self.col)))


def pretty_json(data):
    print(json.dumps(data, sort_keys=True,
                     indent=4, separators=(',', ': ')))


def pretty_rule(rule_str):
    # remove line breaks
    rule_str = ''.join(
        [line.strip() for line in rule_str.strip().splitlines()])

    head_and_body = rule_str.split(':-')

    # drop empty body
    head_and_body = [item.strip()
                     for item in head_and_body if len(item.strip()) > 0]

    head = head_and_body[0]
    if len(head_and_body) == 1:
        return head
    else:
        body = head_and_body[1]
        # split the literal by spliting on ')'
        body_list = body.split(')')
        body_list = body_list[:-1]  # drop part behind the final ')'

        new_body_list = []
        for literal in body_list:
            # remove commas between literals
            if literal[0] == ',':
                literal = literal[1:]
            # add back the ')', also add an indent
            new_body_list.append('  ' + literal.strip() + ')')

        pretty_rule_str = head + " :-\n" + ",\n".join(new_body_list)
        return pretty_rule_str



# Copyright 2017 VMware Inc. All rights reserved.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from congress.db import db_library_policies
from congress.tests import base


class TestDbLibraryPolicies(base.SqlTestCase):

    def setUp(self):
        super(TestDbLibraryPolicies, self).setUp()
        db_library_policies.delete_policies()  # delete preloaded policies

    def test_add_policy_no_name(self):
        self.assertRaises(
            KeyError, db_library_policies.add_policy, {'rules': []})

    def test_add_policy_no_rules(self):
        self.assertRaises(KeyError, db_library_policies.add_policy,
                          {'name': 'policy1'})

    def test_add_policy(self):
        res = db_library_policies.add_policy(
            {'name': 'policy1', 'abbreviation': 'abbr', 'kind': 'database',
             'description': 'descrip', 'rules': [{'rule': 'p(x) :- q(x)',
                                                  'comment': 'test comment',
                                                  'name': 'testname'}]})
        self.assertEqual(res.to_dict(include_rules=True),
                         {'id': res['id'],
                          'abbreviation': 'abbr',
                          'kind': 'database',
                          'name': 'policy1',
                          'description': 'descrip',
                          'rules': [{'rule': 'p(x) :- q(x)',
                                     'comment': 'test comment',
                                     'name': 'testname'}]})

    def test_add_policy_duplicate(self):
        db_library_policies.add_policy(
            {'name': 'policy1', 'abbreviation': 'abbr', 'kind': 'database',
             'description': 'descrip', 'rules': []})
        self.assertRaises(
            KeyError, db_library_policies.add_policy,
            {'name': 'policy1', 'rules': [{'rule': 'p(x) :- q(x)',
                                           'comment': 'test comment',
                                           'name': 'testname'}]})

    def test_get_policy_empty(self):
        res = db_library_policies.get_policies()
        self.assertEqual(res, [])

        self.assertRaises(KeyError, db_library_policies.get_policy,
                          'nosuchpolicy')

        self.assertRaises(KeyError, db_library_policies.get_policy_by_name,
                          'nosuchpolicy')

    def test_create_get_policy(self):
        policy1 = db_library_policies.add_policy(
            {'name': 'policy1', 'abbreviation': 'abbr', 'kind': 'database',
             'description': 'descrip', 'rules': [{'rule': 'p(x) :- q(x)',
                                                  'comment': 'test comment',
                                                  'name': 'testname'}]})

        res = db_library_policies.get_policies()
        self.assertEqual([p.to_dict(include_rules=True) for p in res],
                         [{'id': policy1['id'],
                           'abbreviation': 'abbr',
                           'kind': 'database',
                           'name': 'policy1',
                           'description': 'descrip',
                           'rules': [{'rule': 'p(x) :- q(x)',
                                      'comment': 'test comment',
                                      'name': 'testname'}]}])

        res = db_library_policies.get_policy(policy1['id'])
        self.assertEqual(res.to_dict(include_rules=True),
                         {'id': policy1['id'],
                          'abbreviation': 'abbr',
                          'kind': 'database',
                          'name': 'policy1',
                          'description': 'descrip',
                          'rules': [{'rule': 'p(x) :- q(x)',
                                     'comment': 'test comment',
                                     'name': 'testname'}]})

        res = db_library_policies.get_policy_by_name(policy1['name'])
        self.assertEqual(res.to_dict(include_rules=True),
                         {'id': policy1['id'],
                          'abbreviation': 'abbr',
                          'kind': 'database',
                          'name': 'policy1',
                          'description': 'descrip',
                          'rules': [{'rule': 'p(x) :- q(x)',
                                     'comment': 'test comment',
                                     'name': 'testname'}]})

        self.assertRaises(KeyError, db_library_policies.get_policy,
                          'no_such_policy')

        self.assertRaises(KeyError, db_library_policies.get_policy_by_name,
                          'no_such_policy')

    def test_delete_policy(self):
        db_library_policies.delete_policy('policy1')

        policy1 = db_library_policies.add_policy(
            {'name': 'policy1', 'abbreviation': 'abbr', 'kind': 'database',
             'description': 'descrip', 'rules': [{'rule': 'p(x) :- q(x)',
                                                  'comment': 'test comment',
                                                  'name': 'testname'}]})

        policy2 = db_library_policies.add_policy(
            {'name': 'policy2', 'abbreviation': 'abbr', 'kind': 'database',
             'description': 'descrip', 'rules': [{'rule': 'p(x) :- q(x)',
                                                  'comment': 'test comment',
                                                  'name': 'testname'}]})

        res = db_library_policies.get_policies()
        self.assertEqual(len(res), 2)

        db_library_policies.delete_policy('no_such_policy')

        res = db_library_policies.get_policies()
        self.assertEqual(len(res), 2)

        db_library_policies.delete_policy(policy1['id'])

        res = db_library_policies.get_policies()
        self.assertEqual(len(res), 1)

        db_library_policies.delete_policy(policy2['id'])

        res = db_library_policies.get_policies()
        self.assertEqual(len(res), 0)

    def test_delete_policies(self):
        db_library_policies.delete_policies()
        res = db_library_policies.get_policies()
        self.assertEqual(len(res), 0)

        db_library_policies.add_policy(
            {'name': 'policy1', 'abbreviation': 'abbr', 'kind': 'database',
             'description': 'descrip', 'rules': [{'rule': 'p(x) :- q(x)',
                                                  'comment': 'test comment',
                                                  'name': 'testname'}]})

        db_library_policies.add_policy(
            {'name': 'policy2', 'abbreviation': 'abbr', 'kind': 'database',
             'description': 'descrip', 'rules': [{'rule': 'p(x) :- q(x)',
                                                  'comment': 'test comment',
                                                  'name': 'testname'}]})

        db_library_policies.delete_policies()
        res = db_library_policies.get_policies()
        self.assertEqual(len(res), 0)



#
# Copyright (c) 2016 VMware, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
test_congress_haht
----------------------------------

Replicated policy engine high availability tests for `congress` module.
"""
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import shutil
import subprocess
import sys
import tempfile
import time

from oslo_log import log as logging
import requests
import tenacity

from congress.db import api as db
from congress.db import db_policy_rules
from congress.tests import base
from congress.tests import helper


LOG = logging.getLogger(__name__)


class TestCongressHAHT(base.SqlTestCase):

    class client(object):
        version = '/v1'

        def __init__(self, port, host='0.0.0.0'):
            self.host = host
            self.port = port
            self.base_url = 'http://' + host + ':%d' % port

        def url(self, suffix=None):
            if suffix is None:
                return self.base_url
            else:
                return self.base_url + self.version + '/' + suffix

        def get(self, suffix=None):
            return requests.get(self.url(suffix))

        def delete(self, suffix=None):
            return requests.delete(self.url(suffix))

        def post(self, suffix=None, json=None):
            x = requests.post(self.url(suffix), json=json)
            # print("status: %s, text: %s" % (x.status_code, x.text))
            return x

    def setUp(self):
        super(TestCongressHAHT, self).setUp()
        assert sys.executable is not None,\
            'test cannot proceed when sys.executable is None'

        # establish clean starting DB
        self.clean_db()
        shutil.copy(helper.test_path('haht/test.db.clean'),
                    helper.test_path('haht/test.db'))

        self.clients = []
        self.procs = []
        self.outfiles = {}
        self.errfiles = {}

        self.pe1 = self.start_pe(1, 4001)
        self.pe2 = self.start_pe(2, 4002)

    def dump_nodes_logs(self):
        LOG.error('PE1 process output:\n%s' %
                  self.read_output_file(self.outfiles[1]))
        LOG.error('PE2 process output:\n%s' %
                  self.read_output_file(self.outfiles[2]))

    def clean_db(self):
        session = db.get_session()
        with session.begin(subtransactions=True):
            session.query(db_policy_rules.Policy).delete()
            session.query(db_policy_rules.PolicyRule).delete()

    def start_pe(self, num, port):
        self.outfiles[num] = tempfile.NamedTemporaryFile(
            mode='a+', suffix='.out',
            prefix='congress-pe%d-%d-' % (num, port),
            dir='/tmp')

        self.errfiles[num] = tempfile.NamedTemporaryFile(
            mode='a+', suffix='.err',
            prefix='congress-pe%d-%d-' % (num, port),
            dir='/tmp')

        args = [sys.executable,
                'bin/congress-server',
                '--node-id',
                'node_%d' % num,
                '--api',
                '--policy-engine',
                '--config-file',
                'congress/tests/etc/congress.conf.test.ha_pe%d' % num]
        pe = subprocess.Popen(args,
                              stdout=self.outfiles[num],
                              stderr=self.outfiles[num],
                              cwd=helper.root_path())
        self.addCleanup(pe.kill)
        pe = self.client(port)
        try:
            helper.retry_check_function_return_value(
                lambda: pe.get().status_code, 200)
        except tenacity.RetryError:
            out = self.read_output_file(self.outfiles[num])
            LOG.error('PE%d failed to start. Process output:\n%s' % (num, out))
            raise
        return pe

    def read_output_file(self, file):
        file.flush()
        file.seek(0)
        return ''.join(file.readlines())

    def tail(self, thing, length=20):
        lines = thing.split('\n')
        return '\n'.join(lines[-length:])

    def test_policy_create_delete(self):
        # create policy alice in PE1
        self.assertEqual(self.pe1.post(
            suffix='policies', json={'name': 'alice'}).status_code, 201)
        # check policy alice in PE1
        self.assertEqual(self.pe1.get('policies/alice').status_code, 200)
        # check policy alice in PE2
        helper.retry_check_function_return_value(
            lambda: self.pe2.get('policies/alice').status_code, 200)
        # create policy bob in PE2
        self.assertEqual(self.pe2.post(
            suffix='policies', json={'name': 'bob'}).status_code, 201)
        # check policy bob in PE2
        self.assertEqual(self.pe2.get('policies/bob').status_code, 200)
        # check policy bob in PE1
        helper.retry_check_function_return_value(
            lambda: self.pe1.get('policies/bob').status_code, 200)

        # check policy listings
        self.assertEqual(len(self.pe1.get('policies').json()['results']), 4)
        self.assertEqual(len(self.pe2.get('policies').json()['results']), 4)

        # delete policy alice in PE2, and check deleted on both PE
        self.assertEqual(self.pe2.delete('policies/alice').status_code, 200)
        self.assertEqual(self.pe2.get('policies/alice').status_code, 404)
        helper.retry_check_function_return_value(
            lambda: self.pe1.get('policies/alice').status_code, 404)

        # delete policy bob in PE2, and check deleted on both PE
        self.assertEqual(self.pe2.delete('policies/bob').status_code, 200)
        self.assertEqual(self.pe2.get('policies/bob').status_code, 404)
        helper.retry_check_function_return_value(
            lambda: self.pe1.get('policies/bob').status_code, 404)

    def test_policy_rule_crud(self):
        try:
            # create policy alice in PE1
            self.assertEqual(self.pe1.post(
                suffix='policies', json={'name': 'alice'}).status_code, 201)
            # add rule to PE1
            j = {'rule': 'p(x) :- q(x)', 'name': 'rule1'}
            self.assertEqual(self.pe1.post(
                suffix='policies/alice/rules', json=j).status_code, 201)
            self.assertEqual(
                self.pe1.get('policies/alice/rules').status_code, 200)
            self.assertEqual(
                len(self.pe1.get('policies/alice/rules').
                    json()['results']), 1)
            # retry necessary because of synchronization
            helper.retry_check_function_return_value(
                lambda: len(self.pe2.get('policies/alice/rules').
                            json()['results']), 1)
            # add rule to PE2
            j = {'rule': 'q(1)', 'name': 'rule2'}
            self.assertEqual(self.pe2.post(
                suffix='policies/alice/rules', json=j).status_code, 201)
            # check 2 rule in each pe
            self.assertEqual(len(
                self.pe2.get('policies/alice/rules').json()['results']), 2)
            self.assertEqual(len(
                self.pe1.get('policies/alice/rules').json()['results']), 2)

            # grab rule IDs
            rules = self.pe2.get('policies/alice/rules').json()['results']
            id1 = next(x['id'] for x in rules if x['name'] == 'rule1')
            id2 = next(x['id'] for x in rules if x['name'] == 'rule2')

            # show rules by id
            self.assertEqual(
                self.pe1.get('policies/alice/rules/%s' % id1).status_code, 200)
            self.assertEqual(
                self.pe2.get('policies/alice/rules/%s' % id1).status_code, 200)
            self.assertEqual(
                self.pe1.get('policies/alice/rules/%s' % id2).status_code, 200)
            self.assertEqual(
                self.pe2.get('policies/alice/rules/%s' % id2).status_code, 200)

            # list tables
            self.assertEqual(len(
                self.pe1.get('policies/alice/tables').json()['results']), 2)
            self.assertEqual(len(
                self.pe2.get('policies/alice/tables').json()['results']), 2)

            # show tables
            self.assertEqual(
                self.pe1.get('policies/alice/tables/p').status_code, 200)
            self.assertEqual(
                self.pe2.get('policies/alice/tables/p').status_code, 200)
            self.assertEqual(
                self.pe1.get('policies/alice/tables/q').status_code, 200)
            self.assertEqual(
                self.pe2.get('policies/alice/tables/q').status_code, 200)

            # delete from PE1 and check both have 1 rule left
            self.assertEqual(self.pe1.delete(
                suffix='policies/alice/rules/%s' % id1).status_code, 200)
            self.assertEqual(
                len(self.pe1.get('policies/alice/rules').
                    json()['results']), 1)
            self.assertEqual(
                len(self.pe2.get('policies/alice/rules').
                    json()['results']), 1)
            # delete from PE2 and check both have 0 rules left
            self.assertEqual(self.pe2.delete(
                suffix='policies/alice/rules/%s' % id2).status_code, 200)
            self.assertEqual(
                len(self.pe1.get('policies/alice/rules').
                    json()['results']), 0)
            self.assertEqual(
                len(self.pe2.get('policies/alice/rules').
                    json()['results']), 0)
        except Exception:
            self.dump_nodes_logs()
            raise

    def test_conflicting_policy_create_delete(self):
        try:
            # create policy alice in PE1
            self.assertEqual(self.pe1.post(
                suffix='policies', json={'name': 'alice'}).status_code, 201)
            self.assertEqual(self.pe2.post(
                suffix='policies', json={'name': 'alice'}).status_code, 409)

            # create policy bob in PE1
            self.assertEqual(self.pe1.post(
                suffix='policies', json={'name': 'bob'}).status_code, 201)
            self.assertEqual(self.pe2.delete(
                suffix='policies/bob').status_code, 200)
            self.assertEqual(self.pe1.post(
                suffix='policies', json={'name': 'bob'}).status_code, 201)
        except Exception:
            LOG.error('PE1 process output:\n%s' %
                      self.read_output_file(self.outfiles[1]))
            LOG.error('PE2 process output:\n%s' %
                      self.read_output_file(self.outfiles[2]))
            raise

    def test_policy_rule_create_delete(self):
        try:
            # create policy alice in PE1
            self.assertEqual(self.pe1.post(
                suffix='policies', json={'name': 'alice'}).status_code, 201)
            # add rule to PE1 (retry since 500 on first attempt)
            j = {'rule': 'p(x) :- q(x)', 'name': 'rule1'}
            self.assertEqual(self.pe1.post(
                suffix='policies/alice/rules', json=j).status_code, 201)
            self.assertEqual(
                self.pe1.get('policies/alice/rules').status_code, 200)
            self.assertEqual(
                len(self.pe1.get('policies/alice/rules').
                    json()['results']), 1)
            time.sleep(10)  # wait for sync before reading from PE2
            self.assertEqual(
                len(self.pe2.get('policies/alice/rules').
                    json()['results']), 1)
            # add rule to PE2
            j = {'rule': 'q(1)', 'name': 'rule2'}
            self.assertEqual(self.pe2.post(
                suffix='policies/alice/rules', json=j).status_code, 201)
            # check 2 rule in each pe
            self.assertEqual(len(
                self.pe2.get('policies/alice/rules').json()['results']), 2)
            self.assertEqual(len(
                self.pe1.get('policies/alice/rules').json()['results']), 2)
            # grab rule IDs
            rules = self.pe2.get('policies/alice/rules').json()['results']
            id1 = next(x['id'] for x in rules if x['name'] == 'rule1')
            id2 = next(x['id'] for x in rules if x['name'] == 'rule2')
            # delete from PE1 and check both have 1 rule left
            self.assertEqual(self.pe1.delete(
                suffix='policies/alice/rules/%s' % id1).status_code, 200)
            self.assertEqual(
                len(self.pe1.get('policies/alice/rules').
                    json()['results']), 1)
            self.assertEqual(
                len(self.pe2.get('policies/alice/rules').
                    json()['results']), 1)
            # delete from PE2 and check both have 0 rules left
            self.assertEqual(self.pe2.delete(
                suffix='policies/alice/rules/%s' % id2).status_code, 200)
            self.assertEqual(
                len(self.pe1.get('policies/alice/rules').
                    json()['results']), 0)
            self.assertEqual(
                len(self.pe2.get('policies/alice/rules').
                    json()['results']), 0)
        except Exception:
            self.dump_nodes_logs()
            raise

    def test_policy_rule_create_delete_interference(self):
        try:
            # create policy alice in PE1
            self.assertEqual(self.pe1.post(
                suffix='policies', json={'name': 'alice'}).status_code, 201)
            j = {'rule': 'p(x) :- q(x)', 'name': 'rule1'}

            rule_create_res = self.pe2.post(
                suffix='policies/alice/rules', json=j)
            self.assertEqual(rule_create_res.status_code, 201)
            rule_id = rule_create_res.json()['id']
            self.assertEqual(self.pe1.delete(
                suffix='policies/alice/rules/%s' % rule_id).status_code, 200)

            self.assertEqual(self.pe2.post(
                suffix='policies/alice/rules', json=j).status_code, 201)

        except Exception:
            self.dump_nodes_logs()
            raise

    def test_policy_rule_duplicate(self):
        try:
            # create policy alice in PE1
            self.assertEqual(self.pe1.post(
                suffix='policies', json={'name': 'alice'}).status_code, 201)
            j = {'rule': 'p(x) :- q(x)', 'name': 'rule1'}

            self.assertEqual(self.pe2.post(
                suffix='policies/alice/rules', json=j).status_code, 201)

            self.assertEqual(self.pe1.post(
                suffix='policies/alice/rules', json=j).status_code, 409)

            self.assertEqual(self.pe2.post(
                suffix='policies/alice/rules', json=j).status_code, 409)

            self.assertEqual(
                self.pe1.get('policies/alice/rules').status_code, 200)
            self.assertLessEqual(
                len(self.pe1.get('policies/alice/rules').json()['results']),
                1)
            self.assertEqual(
                len(self.pe2.get('policies/alice/rules').
                    json()['results']), 1)
        except Exception:
            self.dump_nodes_logs()
            raise

    def test_policy_rule_recursion(self):
        try:
            # create policy alice in PE1
            self.assertEqual(self.pe1.post(
                suffix='policies', json={'name': 'alice'}).status_code, 201)
            r1 = {'rule': 'p(x) :- q(x)', 'name': 'rule1'}
            r2 = {'rule': 'q(x) :- p(x)', 'name': 'rule2'}

            self.assertEqual(self.pe2.post(
                suffix='policies/alice/rules', json=r1).status_code, 201)

            self.assertEqual(self.pe1.post(
                suffix='policies/alice/rules', json=r2).status_code, 400)

            self.assertEqual(self.pe2.post(
                suffix='policies/alice/rules', json=r2).status_code, 400)

            self.assertEqual(
                self.pe1.get('policies/alice/rules').status_code, 200)
            self.assertLessEqual(
                len(self.pe1.get('policies/alice/rules').json()['results']),
                1)
            self.assertEqual(
                len(self.pe2.get('policies/alice/rules').
                    json()['results']), 1)
        except Exception:
            self.dump_nodes_logs()
            raise

    def test_policy_rule_schema_mismatch(self):
        try:
            # create policy alice in PE1
            self.assertEqual(self.pe1.post(
                suffix='policies', json={'name': 'alice'}).status_code, 201)
            r1 = {'rule': 'p(x) :- q(x)', 'name': 'rule1'}
            r2 = {'rule': 'p(x) :- q(x, x)', 'name': 'rule2'}

            self.assertEqual(self.pe2.post(
                suffix='policies/alice/rules', json=r1).status_code, 201)

            self.assertEqual(self.pe1.post(
                suffix='policies/alice/rules', json=r2).status_code, 400)

            self.assertEqual(self.pe2.post(
                suffix='policies/alice/rules', json=r2).status_code, 400)

            self.assertEqual(
                self.pe1.get('policies/alice/rules').status_code, 200)
            self.assertLessEqual(
                len(self.pe1.get('policies/alice/rules').json()['results']),
                1)
            self.assertEqual(
                self.pe2.get('policies/alice/rules').status_code, 200)
            self.assertEqual(
                len(self.pe2.get('policies/alice/rules').
                    json()['results']), 1)
        except Exception:
            self.dump_nodes_logs()
            raise

    def test_policy_rule_evaluation(self):
        try:
            # create policy alice in PE1
            self.assertEqual(self.pe1.post(
                suffix='policies', json={'name': 'alice'}).status_code, 201)
            # add rule to PE1
            j = {'rule': 'p(x) :- q(x)', 'name': 'rule0'}
            res = self.pe1.post(
                suffix='policies/alice/rules', json=j)
            self.assertEqual(res.status_code, 201)
            r_id = res.json()['id']

            # add data to PE1
            j = {'rule': '  q(   1   )   ', 'name': 'rule1'}
            res = self.pe1.post(
                suffix='policies/alice/rules', json=j)
            self.assertEqual(res.status_code, 201)
            q1_id = res.json()['id']

            # # add data to PE2
            j = {'rule': '  q   (     2  )   ', 'name': 'rule2'}
            self.assertEqual(self.pe2.post(
                suffix='policies/alice/rules', json=j).status_code, 201)

            # eval on PE1
            helper.retry_check_function_return_value_table(
                lambda: [x['data'] for x in
                         self.pe1.get('policies/alice/tables/p/rows').json()[
                             'results']],
                [[1], [2]])

            # eval on PE2
            helper.retry_check_function_return_value_table(
                lambda: [x['data'] for x in
                         self.pe2.get('policies/alice/tables/p/rows').json()[
                             'results']],
                [[1], [2]])

            self.assertEqual(self.pe1.delete(
                suffix='policies/alice/rules/%s' % q1_id).status_code, 200)

            # eval on PE1
            helper.retry_check_function_return_value_table(
                lambda: [x['data'] for x in
                         self.pe1.get('policies/alice/tables/p/rows').json()[
                             'results']],
                [[2]])

            # eval on PE2
            helper.retry_check_function_return_value_table(
                lambda: [x['data'] for x in
                         self.pe2.get('policies/alice/tables/p/rows').json()[
                             'results']],
                [[2]])

            self.assertEqual(self.pe2.delete(
                suffix='policies/alice/rules/%s' % r_id).status_code, 200)
            helper.retry_check_function_return_value(lambda: self.pe1.get(
                'policies/alice/tables/p/rows').status_code, 404)
            helper.retry_check_function_return_value(lambda: self.pe2.get(
                'policies/alice/tables/p/rows').status_code, 404)

        except Exception:
            self.dump_nodes_logs()
            raise



# Copyright (c) 2015 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from congress.datasources import datasource_driver


class PerformanceTestDriver(datasource_driver.PollingDataSourceDriver):
    TABLE = 'p'

    # This is the most common per-value translator, so define it once here.
    value_trans = {'translation-type': 'VALUE'}

    p_translator = {
        'translation-type': 'HDICT',
        'table-name': TABLE,
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'field1', 'translator': value_trans},
             {'fieldname': 'field2', 'translator': value_trans},
             {'fieldname': 'field3', 'translator': value_trans},
             {'fieldname': 'field4', 'translator': value_trans},
             {'fieldname': 'field5', 'translator': value_trans},
             {'fieldname': 'field6', 'translator': value_trans})}

    def __init__(self, name='', args=None):
        # if args is None:
        #     args = self._empty_openstack_credentials()
        super(PerformanceTestDriver, self).__init__(name, args)
        self.client_data = None
        self.register_translator(PerformanceTestDriver.p_translator)
        self._init_end_start_poll()

    @staticmethod
    def get_datasource_info():
        result = {}
        result['id'] = 'performance'
        result['description'] = 'Datasource driver used for perf tests'
        # result['config'] = ds_utils.get_openstack_required_config()
        # result['config']['api_version'] = constants.OPTIONAL
        result['secret'] = ['password']
        return result

    def update_from_datasource(self):
        if self.client_data is not None:
            self.state = {}
            row_data = self.convert_objs(self.client_data, self.p_translator)
            self.state[self.TABLE] = set()
            for table, row in row_data:
                assert table == self.TABLE
                self.state[table].add(row)



# Copyright (c) 2016 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import copy
import six  # thirdparty import first because needed in import of Queue/queue
import time
if six.PY2:
    import Queue as queue_package
else:
    import queue as queue_package

from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils as json

from congress import exception

LOG = logging.getLogger(__name__)


class DataServiceInfo(object):
    """Metadata for DataService on the DSE.

    Attributes:
    - service_id: The ID of the service.
    - node_id: The ID of the node the service is running on.
    - published_tables: List of table IDs published by this service.
    - subscribed_tables: List of table IDs this service subscribes to.
    - rpc_endpoints_info: List of RPC endpoints exposed by this service.
    """
    MARSHALL_ATTRS = set(['service_id', 'node_id', 'published_tables',
                          'subscribed_tables', 'rpc_endpoints_info'])

    def __init__(self, service_id=None, node_id=None, published_tables=None,
                 subscribed_tables=None, rpc_endpoints_info=None):
        self.service_id = service_id
        self.node_id = node_id
        self.published_tables = published_tables or []
        self.subscribed_tables = subscribed_tables or []
        self.rpc_endpoints_info = rpc_endpoints_info or []

    def __str__(self):
        return self.__class__.__name__ + "<%s>" % self.service_id

    def __repr__(self):
        return self.__class__.__name__ + "(%s)" % dict.__repr__(vars(self))

    @classmethod
    def from_json(cls, json_string):
        return json.loads(json_string, object_hook=cls.from_dict)

    @classmethod
    def from_dict(cls, raw_dict):
        provided_keys = set(raw_dict.keys())
        if provided_keys != cls.MARSHALL_ATTRS:
            missing = cls.MARSHALL_ATTRS - provided_keys
            malformed = provided_keys - cls.MARSHALL_ATTRS
            msg = "Cannot construct %s from input:" % cls.__name__
            if missing:
                msg += " Missing keys: %s" % list(missing)
            if malformed:
                msg += " Malformed keys: %s" % list(malformed)
            raise KeyError(msg)
        ret = DataServiceInfo()
        for n in cls.MARSHALL_ATTRS:
            setattr(ret, n, raw_dict[n])
        return ret

    def to_dict(self):
        return dict([(k, getattr(self, k)) for k in self.MARSHALL_ATTRS])

    def to_json(self):
        return json.dumps(self.to_dict())


class DataService(object):
    """A unit of data and business logic that interfaces with the DSE.

    A DataService may publish tables, subscribe to tables, and/or expose
    RPCs on the DSE.  DataService instances are bound to a DseNode which is
    used for all inter-service communication.

    Attributes:

    - service_id: A unique ID of the service.
    - _published_tables_with_subscriber: A set of tables published by self
      that has subscribers
    """

    # TODO(pballand): make default methods for pub/subscribed tables
    def __init__(self, service_id):
        # Note(ekcs): temporary setting to disable use of diffs and sequencing
        #   to avoid muddying the process of a first dse2 system test.
        self.service_id = service_id
        self.node = None
        self._rpc_server = None
        self._target = None
        self._rpc_endpoints = [DataServiceEndPoints(self)]
        self._running = False
        self._published_tables_with_subscriber = set()

        # data structures for sequenced data updates for reliable pub-sub
        # msg queues for msgs to be processed
        # each queue is a priority queue prioritizing lowest seqnum
        self.msg_queues = {}  # {publisher -> {table -> msg queue}}
        # oldest queued msg receipt time
        self.oldest_queue_times = {}  # {publisher -> {table -> receipt time}}
        # last received & processed seqnum
        self.receiver_seqnums = {}  # {publisher -> {table -> seqnum}}
        # last sent seqnum
        self.sender_seqnums = {}  # {table -> seqnum}
        # last published data
        self._last_published_data = {}  # {table -> data}

        # custom functions to execute on each heartbeat
        # must be 0-ary function
        # {id -> function}
        self.heartbeat_callbacks = {}

    def add_rpc_endpoint(self, endpt):
        self._rpc_endpoints.append(endpt)

    def rpc_endpoints(self):
        """Return list of RPC endpoint objects to be exposed for this service.

        A DataService may include zero or more RPC endpoints to be exposed
        by the DseNode.  Each endpoint object must be compatible with the
        oslo.messaging RPC Server.
        """
        return self._rpc_endpoints

    @property
    def status(self):
        return "OK"

    @property
    def info(self):
        # TODO(pballand): populate rpc_endpoints_info from rpc_endpoints
        return DataServiceInfo(
            service_id=self.service_id,
            node_id=self.node.node_id if self.node else None,
            published_tables=None,
            subscribed_tables=self._published_tables_with_subscriber,
            rpc_endpoints_info=None,
            )

    def start(self):
        """Start the DataService.

        This method is called by a DseNode before any RPCs are invoked.
        """
        assert self.node is not None
        if not self._running:
            self._rpc_server.start()
            self._running = True

    def stop(self):
        """Stop the DataService.

        This method is called by a DseNode when the DataService instance is
        no longer needed.  No RPCs will invoked on stopped DataServices.
        """
        assert self.node is not None
        self._rpc_server.stop()
        self._running = False

    # Note(thread-safety): blocking function
    def wait(self):
        """Wait for processing to complete.

        After a call to stop(), the DataService may have some outstanding work
        that has not yet completed.  The wait() method blocks until all
        DataService processing is complete.
        """
        assert self.node is not None
        self._rpc_server.wait()

    # Note(thread-safety): blocking function
    def rpc(self, service, action, kwargs=None, timeout=None, local=False,
            retry=None):
        return self.node.invoke_service_rpc(
            service, action, kwargs, timeout=timeout, local=local, retry=retry)

    # Will be removed once the reference of node exists in api
    # Note(thread-safety): blocking function
    def get_datasources(self, filter_secret=False):
        return self.node.get_datasources(filter_secret)

    def is_valid_service(self, service_id):
        return self.node.is_valid_service(service_id)

    # Will be removed once the reference of node exists in api
    # Note(thread-safety): blocking function
    def get_datasource(self, datasource_id):
        return self.node.get_datasource(datasource_id)

    # Will be removed once the reference of node exists in api
    # Note(thread-safety): blocking function
    def get_drivers_info(self, *args):
        return self.node.get_drivers_info(*args)

    # Will be removed once the reference of node exists in api
    # Note(thread-safety): blocking function
    def get_driver_info(self, *args):
        return self.node.get_driver_info(*args)

    # Will be removed once the reference of node exists in api
    # Note(thread-safety): blocking function
    def get_driver_schema(self, *args):
        return self.node.get_driver_schema(*args)

    # Will be removed once the reference of node exists in api
    # Note(thread-safety): blocking function
    def make_datasource_dict(self, *args, **kwargs):
        return self.node.make_datasource_dict(*args, **kwargs)

    # Note(thread-safety): blocking function
    def publish(self, table, data, use_snapshot=False):
        def get_differential_and_set_last_published_data():
            if table in self._last_published_data:
                to_add = list(
                    set(data) - set(self._last_published_data[table]))
                to_del = list(
                    set(self._last_published_data[table]) - set(data))
                self._last_published_data[table] = data
            else:
                self._last_published_data[table] = data
                to_add = copy.copy(data)
                to_del = []
            return [to_add, to_del]

        def increment_get_seqnum():
            if table not in self.sender_seqnums:
                self.sender_seqnums[table] = 0
            else:
                self.sender_seqnums[table] = self.sender_seqnums[table] + 1
            return self.sender_seqnums[table]

        if not use_snapshot:
            data = get_differential_and_set_last_published_data()
            if len(data[0]) == 0 and len(data[1]) == 0:
                return

        seqnum = increment_get_seqnum()
        self.node.publish_table_sequenced(
            self.service_id, table, data, use_snapshot, seqnum)

    # Note(thread-safety): blocking function
    def subscribe(self, service, table):
        try:
            # Note(thread-safety): blocking call
            (seqnum, data) = self.node.subscribe_table(
                self.service_id, service, table)
            self.receive_data_sequenced(
                service, table, data, seqnum, is_snapshot=True)
        except exception.NotFound:
            LOG.info("Service '%s' unresponsive. '%s:%s' subscribed but data "
                     "not yet initialized.", service, service, table)

    def unsubscribe(self, service, table):
        # Note(thread-safety): it is important to make sure there are no
        #             blocking calls in modifying the msg_queues and related
        #             data structures.
        #             Otherwise, concurrency bugs will likely occur when the
        #             periodic task _check_resub_all interrupts and modifies
        #             the same data structures.
        self.node.unsubscribe_table(self.service_id, service, table)
        self._clear_msg_queue(service, table)
        self._clear_receiver_seqnum(service, table)

    def _clear_msg_queue(self, publisher, table):
        # Note(thread-safety): it is important to make sure there are no
        #             blocking calls in modifying the msg_queues and related
        #             data structures.
        #             Otherwise, concurrency bugs will likely occur when the
        #             periodic task _check_resub_all interrupts and modifies
        #             the same data structures.
        if publisher in self.msg_queues:
            if table in self.msg_queues[publisher]:
                del self.msg_queues[publisher][table]
                del self.oldest_queue_times[publisher][table]

    def _clear_receiver_seqnum(self, publisher, table):
        # Note(thread-safety): it is important to make sure there are no
        #             blocking calls in modifying the msg_queues and related
        #             data structures.
        #             Otherwise, concurrency bugs will likely occur when the
        #             periodic task _check_resub_all interrupts and modifies
        #             the same data structures.
        if publisher in self.receiver_seqnums:
            if table in self.receiver_seqnums[publisher]:
                del self.receiver_seqnums[publisher][table]

    def receive_data_sequenced(
            self, publisher, table, data, seqnum, is_snapshot=False,
            receipt_time=None):
        """Method called when sequenced publication data arrives."""
        # Note(thread-safety): it is important to make sure there are no
        #             blocking calls in modifying the msg_queues and related
        #             data structures.
        #             Otherwise, concurrency bugs will likely occur when the
        #             periodic task _check_resub_all interrupts and modifies
        #             the same data structures.
        # TODO(ekcs): allow opting out of sequenced processing (per table)
        def set_seqnum():
            if publisher not in self.receiver_seqnums:
                self.receiver_seqnums[publisher] = {}
            self.receiver_seqnums[publisher][table] = seqnum

        def clear_msg_queue():
            self._clear_msg_queue(publisher, table)

        def add_to_msg_queue():
            if publisher not in self.msg_queues:
                self.msg_queues[publisher] = {}
                self.oldest_queue_times[publisher] = {}
            if table not in self.msg_queues[publisher]:
                self.msg_queues[publisher][table] = \
                    queue_package.PriorityQueue()
                # set oldest queue time on first msg only, others are newer
                self.oldest_queue_times[publisher][table] = receipt_time
            self.msg_queues[publisher][table].put_nowait(
                (seqnum, is_snapshot, data, receipt_time))
            assert self.msg_queues[publisher][table].qsize() > 0

        def process_queued_msg():
            def update_oldest_time():
                '''Set oldest time to the receipt time of oldest item in queue.

                Called after removing the previous oldest item from a queue.
                If queue is empty, corresponding oldest time is set to None.
                '''
                try:
                    # remove and then add back to priority queue to get the
                    # receipt time of the next oldest message
                    # (peek not available in python standard library queues)
                    s, i, d, t = self.msg_queues[publisher][table].get_nowait()
                    self.msg_queues[publisher][table].put_nowait((s, i, d, t))
                    self.oldest_queue_times[publisher][table] = t
                except queue_package.Empty:
                    # set oldest time to None if queue empty
                    self.oldest_queue_times[publisher][table] = None

            try:
                s, i, d, t = self.msg_queues[publisher][table].get_nowait()
                update_oldest_time()
                self.receive_data_sequenced(publisher, table, d, s, i, t)
            except queue_package.Empty:
                pass
            except KeyError:
                pass

        if receipt_time is None:
            receipt_time = time.time()

        # if no seqnum process immediately
        if seqnum is None:
            self.receive_data(publisher, table, data, is_snapshot)

        # if first data update received on this table
        elif (publisher not in self.receiver_seqnums or
                table not in self.receiver_seqnums[publisher]):
            if is_snapshot:
                # set sequence number and process data
                set_seqnum()
                self.receive_data(publisher, table, data, is_snapshot)
                process_queued_msg()
            else:
                # queue
                add_to_msg_queue()

        # if re-initialization
        elif seqnum == 0:  # initial snapshot or reset
            # set sequence number and process data
            set_seqnum()
            clear_msg_queue()
            self.receive_data(publisher, table, data, is_snapshot)

        else:
            # if seqnum is old, ignore
            if seqnum <= self.receiver_seqnums[publisher][table]:
                process_queued_msg()

            # if seqnum next, process all in sequence
            elif seqnum == self.receiver_seqnums[publisher][table] + 1:
                set_seqnum()
                self.receive_data(publisher, table, data, is_snapshot)
                process_queued_msg()

            # if seqnum future, queue for future
            elif seqnum > self.receiver_seqnums[publisher][table] + 1:
                add_to_msg_queue()

    def receive_data(self, publisher, table, data, is_snapshot=True):
        """Method called when publication data arrives.

           Instances will override this method.
        """
        if is_snapshot:
            data = self.node.to_set_of_tuples(data)
        else:
            data = (self.node.to_set_of_tuples(data[0]),
                    self.node.to_set_of_tuples(data[1]))

        self.last_msg = {}
        self.last_msg['data'] = data
        self.last_msg['publisher'] = publisher
        self.last_msg['table'] = table

        if not hasattr(self, 'receive_data_history'):
            self.receive_data_history = []
        self.receive_data_history.append(self.last_msg)

    def subscription_list(self):
        """Method that returns subscription list.

        It returns list of tuple that represents the service's subscription.
        The tuple forms following format:
        (service_id, table_name).
        """
        result = []
        subscription = self.node.get_subscription(self.service_id)
        for target, tables in subscription.items():
            result.extend([(target, t) for t in tables])
        return result

    def subscriber_list(self):
        """Method that returns subscribers list.

        It returns list of services subscribed to this service data.
        """
        return self.node.get_subscribers(self.service_id)

    def get_last_published_data_with_seqnum(self, table):
        """Method that returns the current seqnum & data for given table."""
        if table not in self.sender_seqnums:
            self.sender_seqnums[table] = 0
            self._last_published_data[table] = self.get_snapshot(table)
        return (self.sender_seqnums[table], self._last_published_data[table])

    def get_snapshot(self, table):
        """Method that returns the current data for the given table.

           Should be overridden.
        """
        raise NotImplementedError(
            "get_snapshot is not implemented in the '%s' class." %
            self.service_id)

    def check_resub_all(self):
        '''Check all subscriptions for long missing update and resubscribe.'''
        def check_resub(publisher, table):
            if (publisher in self.oldest_queue_times and
                table in self.oldest_queue_times[publisher] and
                self.oldest_queue_times[publisher][table] is not None and
                (time.time() - self.oldest_queue_times[publisher][table]
                 > cfg.CONF.dse.time_to_resub)):
                self.unsubscribe(publisher, table)
                self.subscribe(publisher, table)
                return True
            else:
                return False

        copy_oldest_queue_times = copy.copy(self.oldest_queue_times)
        for publisher in copy_oldest_queue_times:
            copy_oldest_queue_times_publisher = copy.copy(
                copy_oldest_queue_times[publisher])
            for table in copy_oldest_queue_times_publisher:
                check_resub(publisher, table)


class DataServiceEndPoints (object):
    def __init__(self, service):
        self.service = service

    def get_snapshot(self, context, table):
        """Function called on a node when an RPC request is sent."""
        try:
            return self.service.get_snapshot(table)
        except AttributeError:
            pass

    def get_last_published_data_with_seqnum(self, context, table):
        """Function called on a node when an RPC request is sent."""
        try:
            return self.service.get_last_published_data_with_seqnum(table)
        except AttributeError:
            pass

    def ping(self, client_ctxt, **args):
        """Echo args"""
        return args



# Copyright (c) 2014 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import


class ResponseObj(object):
    """Allows callers to use dot notation to access a dictionary."""
    def __init__(self, values):
        self.values = values

    def __getattr__(self, name):
        return self.values.get(name)

    def to_dict(self):
        return self.values



#    Copyright 2018 Orange
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Unit tests for typechecker"""

import six

from congress import data_types
from congress.datalog import base as datalog
from congress.datalog import compile as ast
from congress.datalog import nonrecursive
from congress.datalog import ruleset
from congress.tests import base
from congress.z3 import typechecker


def mkc(typ, nullable):
    return {'type': typ, 'nullable': nullable}


class TestMinTypes(base.TestCase):

    def setUp(self):
        try:
            data_types.TypesRegistry.type_class('Enum')
        except KeyError:
            typ = data_types.create_congress_enum_type(
                'Enum', ['a', 'b', 'c'], data_types.Str)
            data_types.TypesRegistry.register(typ)
        super(TestMinTypes, self).setUp()

    def test_not_convertible(self):
        self.assertIsNone(typechecker.min_type('Str', 'Int', False))

    def test_convertible(self):
        self.assertEqual('Enum', typechecker.min_type('Str', 'Enum', False))
        self.assertEqual('Enum', typechecker.min_type('Enum', 'Str', False))

    def test_constrained(self):
        self.assertEqual('Enum', typechecker.min_type('Str', 'Enum', True))
        self.assertIsNone(typechecker.min_type('Enum', 'Str', True))


class TestCellPrimitives(base.TestCase):

    def test_constrain1(self):
        tc = typechecker.Typechecker([], [])
        cell = mkc(None, False)
        tc.work = False
        tc.constrain_type(cell, 'Str')
        self.assertEqual('Str', cell['type'])
        self.assertIs(True, tc.work)

    def test_constrain2(self):
        tc = typechecker.Typechecker([], [])
        cell = mkc('Int', False)
        tc.work = False
        tc.constrain_type(cell, 'Str')
        self.assertEqual('Scalar', cell['type'])
        self.assertIs(True, tc.work)

    def test_constrain3(self):
        tc = typechecker.Typechecker([], [])
        cell = mkc('Str', False)
        tc.work = False
        tc.constrain_type(cell, 'Str')
        self.assertEqual('Str', cell['type'])
        self.assertIs(False, tc.work)

    def test_nullable1(self):
        tc = typechecker.Typechecker([], [])
        cell = mkc('Str', False)
        tc.work = False
        tc.set_nullable(cell)
        self.assertIs(True, cell['nullable'])
        self.assertIs(True, tc.work)

    def test_nullable2(self):
        tc = typechecker.Typechecker([], [])
        cell = mkc('Str', True)
        tc.work = False
        tc.set_nullable(cell)
        self.assertIs(True, cell['nullable'])
        self.assertIs(False, tc.work)

    def test_type_cells1(self):
        tc = typechecker.Typechecker([], [])
        cell1, cell2 = mkc('Str', True), mkc(None, False)
        self.assertIsNone(tc.type_cells(cell1, cell2, False))
        self.assertEqual(mkc('Str', True), cell1)
        self.assertEqual(mkc('Str', True), cell2)
        self.assertIs(True, tc.work)

    def test_type_cells2(self):
        tc = typechecker.Typechecker([], [])
        cell1, cell2 = mkc(None, False), mkc('Str', True)
        self.assertIsNone(tc.type_cells(cell1, cell2, False))
        self.assertEqual(mkc('Str', True), cell1)
        self.assertEqual(mkc('Str', True), cell2)
        self.assertIs(True, tc.work)

    def test_type_cells3(self):
        tc = typechecker.Typechecker([], [])
        cell1, cell2 = mkc(None, False), mkc('Str', True)
        self.assertIsNone(tc.type_cells(cell1, cell2, False))
        self.assertEqual(mkc('Str', True), cell1)
        self.assertEqual(mkc('Str', True), cell2)
        self.assertIs(True, tc.work)

    def test_type_cells4(self):
        tc = typechecker.Typechecker([], [])
        cell1, cell2 = mkc('Str', False), mkc('Str', False)
        self.assertIsNone(tc.type_cells(cell1, cell2, False))
        self.assertEqual(mkc('Str', False), cell1)
        self.assertEqual(mkc('Str', False), cell2)
        self.assertIs(False, tc.work)
        cell1, cell2 = mkc('Str', True), mkc('Str', True)
        self.assertIsNone(tc.type_cells(cell1, cell2, False))
        cell1, cell2 = mkc(None, False), mkc(None, False)
        self.assertIsNone(tc.type_cells(cell1, cell2, False))
        self.assertIs(False, tc.work)

    def test_type_cells5(self):
        tc = typechecker.Typechecker([], [])
        cell1, cell2 = mkc('Int', False), mkc('Str', True)
        self.assertIsNotNone(tc.type_cells(cell1, cell2, False))

    def test_type_constant(self):
        tc = typechecker.Typechecker([], [])

        def check(val, typ, nullable):
            cell = mkc(None, False)
            tc.type_constant(val, cell)
            self.assertEqual(mkc(typ, nullable), cell)

        check(1, 'Int', False)
        check('aaa', 'Str', False)
        check(True, 'Bool', False)
        check(1.3, 'Float', False)
        check(None, None, True)
        check((1, 3), 'Scalar', False)


class MinTheory(nonrecursive.RuleHandlingMixin, datalog.Theory):

    def __init__(self, name, theories):
        super(MinTheory, self).__init__(name=name, theories=theories)
        self.rules = ruleset.RuleSet()
        self.schema = ast.Schema()


class TestTypeChecker(base.TestCase):

    def setUp(self):
        self.world = {}
        t1 = MinTheory('t1', self.world)
        t2 = MinTheory('t2', self.world)
        self.world['t1'] = t1
        self.world['t2'] = t2
        self.rules = ast.parse(
            'l(2). l(3). p(x) :- l(x). q(x,x) :- m(x). '
            'm("a"). k(x) :- t2:f(x). r(y) :- q(x,y).')
        for rule in self.rules:
            t1.insert(rule)
        for rule in ast.parse("f(3)."):
            t2.insert(rule)
        self.t1 = t1
        self.t2 = t2
        super(TestTypeChecker, self).setUp()

    def test_reset(self):
        tc = typechecker.Typechecker([self.t1], self.world)
        tc.reset_types()
        sch1 = self.t1.schema
        for (_, cols) in six.iteritems(sch1.map):
            for col in cols:
                self.assertIs(False, col['nullable'])
                self.assertIsNone(col['type'])

    def test_reset_variables(self):
        tc = typechecker.Typechecker([self.t1], self.world)
        tc.reset_type_environment()
        env = tc.type_env
        self.assertEqual(4, len(env.keys()))
        for variables in six.itervalues(env):
            for (v, cell) in six.iteritems(variables):
                self.assertIn(v, [u'x', u'y'])
                self.assertEqual(mkc(None, False), cell)

    def test_type_facts(self):
        tc = typechecker.Typechecker([self.t1], self.world)
        tc.reset_types()
        tc.reset_type_environment()
        tc.type_facts(self.t1)
        cols1 = self.t1.schema.map['l']
        self.assertEqual(1, len(cols1))
        self.assertEqual('Int', cols1[0]['type'])
        self.assertIs(False, cols1[0]['nullable'])
        cols2 = self.t1.schema.map['m']
        self.assertEqual(1, len(cols2))
        self.assertEqual('Str', cols2[0]['type'])
        self.assertIs(False, cols2[0]['nullable'])

    def test_type_rule(self):
        rule = self.rules[2]
        tc = typechecker.Typechecker([self.t1], self.world)
        tc.reset_types()
        tc.reset_type_environment()
        tc.type_facts(self.t1)
        tc.type_rule(self.t1, rule)
        self.assertEqual(mkc('Int', False), tc.type_env[rule.id]['x'])
        self.assertEqual('Int', self.t1.schema.map['p'][0]['type'])
        self.assertIs(True, tc.work)

    def test_type_all(self):
        tc = typechecker.Typechecker([self.t1], self.world)
        tc.type_all()
        smap = self.t1.schema.map
        self.assertEqual('Int', smap['p'][0]['type'])      # propagation of l
        self.assertEqual('Str', smap['q'][0]['type'])      # propagation of m
        self.assertEqual('Str', smap['q'][1]['type'])
        self.assertEqual('Str', smap['r'][0]['type'])
        self.assertEqual('Scalar', smap['k'][0]['type'])   # prop of ext table



# Copyright (c) 2012 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import json

import sqlalchemy as sa
from sqlalchemy.orm import exc as db_exc

from congress.db import api as db
from congress.db import db_ds_table_data as table_data
from congress.db import model_base
from congress.db import utils as db_utils
from congress import encryption


class Datasource(model_base.BASE, model_base.HasId):
    __tablename__ = 'datasources'

    name = sa.Column(sa.String(255), unique=True)
    driver = sa.Column(sa.String(255))
    config = sa.Column(sa.Text(), nullable=False)
    description = sa.Column(sa.Text(), nullable=True)
    enabled = sa.Column(sa.Boolean, default=True)

    def __init__(self, id_, name, driver, config, description,
                 enabled=True):
        self.id = id_
        self.name = name
        self.driver = driver
        self.config = json.dumps(config)
        self.description = description
        self.enabled = enabled


def _encrypt_secret_config_fields(ds_db_obj, secret_config_fields):
    '''encrypt secret config fields'''
    config = json.loads(ds_db_obj.config)
    if config is None:  # nothing to encrypt
        return ds_db_obj  # return original obj
    if '__encrypted_fields' in config:
        raise Exception('Attempting to encrypt already encrypted datasource '
                        'DB object. This should not occer.')
    for field in secret_config_fields:
        config[field] = encryption.encrypt(config[field])
    config['__encrypted_fields'] = secret_config_fields
    ds_db_obj.config = json.dumps(config)
    return ds_db_obj


def _decrypt_secret_config_fields(ds_db_obj):
    '''de-encrypt previously encrypted secret config fields'''
    config = json.loads(ds_db_obj.config)
    if config is None:
        return ds_db_obj  # return original object
    if '__encrypted_fields' not in config:  # not previously encrypted
        return ds_db_obj  # return original object
    else:
        for field in config['__encrypted_fields']:
            config[field] = encryption.decrypt(config[field])
        del config['__encrypted_fields']
        ds_db_obj.config = json.dumps(config)
        return ds_db_obj


@db_utils.retry_on_db_error
def add_datasource(id_, name, driver, config, description,
                   enabled, session=None, secret_config_fields=None):
    secret_config_fields = secret_config_fields or []
    session = session or db.get_session()
    with session.begin(subtransactions=True):
        datasource = Datasource(
            id_=id_,
            name=name,
            driver=driver,
            config=config,
            description=description,
            enabled=enabled)
        _encrypt_secret_config_fields(datasource, secret_config_fields)
        session.add(datasource)
    return datasource


@db_utils.retry_on_db_error
def delete_datasource(id_, session=None):
    session = session or db.get_session()
    return session.query(Datasource).filter(
        Datasource.id == id_).delete()


@db_utils.retry_on_db_error
def delete_datasource_with_data(id_, session=None):
    session = session or db.get_session()
    with session.begin(subtransactions=True):
        deleted = delete_datasource(id_, session)
        table_data.delete_ds_table_data(id_, session=session)
    return deleted


@db_utils.retry_on_db_error
def get_datasource_name(name_or_id, session=None):
    session = session or db.get_session()
    datasource_obj = get_datasource(name_or_id, session)
    if datasource_obj is not None:
        return datasource_obj.name
    return name_or_id


def get_datasource(name_or_id, session=None):
    db_object = (get_datasource_by_name(name_or_id, session) or
                 get_datasource_by_id(name_or_id, session))

    return db_object


@db_utils.retry_on_db_error
def get_datasource_by_id(id_, session=None):
    session = session or db.get_session()
    try:
        return _decrypt_secret_config_fields(session.query(Datasource).
                                             filter(Datasource.id == id_).
                                             one())
    except db_exc.NoResultFound:
        pass


@db_utils.retry_on_db_error
def get_datasource_by_name(name, session=None):
    session = session or db.get_session()
    try:
        return _decrypt_secret_config_fields(session.query(Datasource).
                                             filter(Datasource.name == name).
                                             one())
    except db_exc.NoResultFound:
        pass


@db_utils.retry_on_db_error
def get_datasources(session=None, deleted=False):
    session = session or db.get_session()
    return [_decrypt_secret_config_fields(ds_obj)
            for ds_obj in session.query(Datasource).all()]



#!/usr/bin/env python
# Copyright (c) 2014 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
import neutronclient.v2_0.client
from oslo_log import log as logging

from congress.datasources import constants
from congress.datasources import datasource_driver
from congress.datasources import datasource_utils as ds_utils

LOG = logging.getLogger(__name__)


class NeutronV2QosDriver(datasource_driver.PollingDataSourceDriver):

    PORTS = 'ports'
    QOS_POLICY_PORT_BINDINGS = 'qos_policy_port_bindings'
    QOS_POLICIES = 'policies'
    QOS_RULES = 'rules'

    # This is the most common per-value translator, so define it once here.
    value_trans = {'translation-type': 'VALUE'}

    ports_qos_policies_translator = {
        'translation-type': 'LIST',
        'table-name': QOS_POLICY_PORT_BINDINGS,
        'parent-key': 'id',
        'parent-col-name': 'port_id',
        'parent-key-desc': 'UUID of port',
        'val-col': 'qos_policy_id',
        'val-col-desc': 'UUID of qos policy',
        'translator': value_trans}

    ports_translator = {
        'translation-type': 'HDICT',
        'table-name': PORTS,
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'id', 'desc': 'UUID of port',
              'translator': value_trans},
             {'fieldname': 'qos_policies',
              'translator': ports_qos_policies_translator})}

    qos_rules_translator = {
        'translation-type': 'HDICT',
        'table-name': QOS_RULES,
        'parent-key': 'id',
        'parent-col-name': 'qos_policy_id',
        'parent-key-desc': 'uuid of qos policy',
        'selector-type': 'DICT_SELECTOR',
        'in-list': True,
        'field-translators':
            ({'fieldname': 'id', 'desc': 'The UUID for the QoS minimum'
                                         'bandwidth rule translator',
              'translator': value_trans},
             {'fieldname': 'min_kbps', 'desc': 'min_kbps bandwidth '
                                               'limit rule',
              'translator': value_trans},
             {'fieldname': 'direction', 'desc': 'minimum bandwidth '
                                                'rule direction',
              'translator': value_trans},
             {'fieldname': 'type', 'desc': 'type of qos rule',
              'translator': value_trans},
             {'fieldname': 'dscp_mark', 'desc': 'mark of the dscp rule',
              'translator': value_trans},
             {'fieldname': 'max_burst_kbps', 'desc': 'max_burst_kps limit',
              'translator': value_trans},
             {'fieldname': 'max_kbps', 'desc': 'max_kps bandwidth limit',
              'translator': value_trans})}

    qos_policy_translator = {
        'translation-type': 'HDICT',
        'table-name': QOS_POLICIES,
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'id', 'desc': 'The UUID for the qos policy',
              'translator': value_trans},
             {'fieldname': 'tenant_id', 'desc': 'Tenant ID',
              'translator': value_trans},
             {'fieldname': 'name', 'desc': 'The qos policy name',
              'translator': value_trans},
             {'fieldname': 'description', 'desc': 'qos policy description',
              'translator': value_trans},
             {'fieldname': 'shared', 'desc': 'The qos policy share',
              'translator': value_trans},
             {'fieldname': 'rules',
              'translator': qos_rules_translator})}

    TRANSLATORS = [ports_translator, qos_policy_translator]

    def __init__(self, name='', args=None):
        super(NeutronV2QosDriver, self).__init__(name, args=args)
        self.creds = args
        session = ds_utils.get_keystone_session(self.creds)
        self.neutron = neutronclient.v2_0.client.Client(session=session)
        self.initialize_update_methods()
        self._init_end_start_poll()

    @staticmethod
    def get_datasource_info():
        result = {}
        result['id'] = 'neutronv2_qos'
        result['description'] = ('Datasource driver that interfaces with QoS '
                                 'extension of '
                                 'OpenStack Networking aka Neutron.')
        result['config'] = ds_utils.get_openstack_required_config()
        result['config']['lazy_tables'] = constants.OPTIONAL
        result['secret'] = ['password']
        return result

    def initialize_update_methods(self):
        ports_method = lambda: self._translate_ports(self.neutron.list_ports())
        self.add_update_method(ports_method, self.ports_translator)
        qos_policy_method = lambda: self._translate_qos_policies(
            self.neutron.list_qos_policies())
        self.add_update_method(qos_policy_method,
                               self.qos_policy_translator)

    @ds_utils.update_state_on_changed(PORTS)
    def _translate_ports(self, obj):
        LOG.debug("ports: %s", obj)
        row_data = NeutronV2QosDriver.convert_objs(obj['ports'],
                                                   self.ports_translator)
        return row_data

    @ds_utils.update_state_on_changed(QOS_POLICIES)
    def _translate_qos_policies(self, obj):
        LOG.debug("qos_policies: %s", obj)
        row_data = NeutronV2QosDriver.convert_objs(obj['policies'],
                                                   self.qos_policy_translator)
        return row_data



#
# Copyright (c) 2017 Orange.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
"""Tests for the config validator agent."""

from os import path

import json
import mock
import six

from oslo_config import cfg
from oslo_log import log as logging

from congress.cfg_validator.agent import agent
from congress.cfg_validator.agent import opts
from congress import opts as opts_congress
from congress.tests import base

# pylint: disable=protected-access

LOG = logging.getLogger(__name__)

NAMESPACE_FILE_CONTENT = """[DEFAULT]
output_file = etc/svc.conf.sample
wrap_width = 79
namespace = congress
namespace = congress-agent
"""

CONF_FILE1_CONTENT = """
[agent]
version:v0
"""

CONF_FILE2_CONTENT = """
[agent]
host:h0
"""

CONF_AGENT_CONTENT = """
[agent]
host:hhh
version:vvv
services: svc: {svc1.conf: svc.tpl}
"""

TEMPLATES = {
    "tpl1": agent.Template("TPL1", "out"),
    "tpl2": agent.Template("TPL2", "out")}
CONFIGS = {
    "cfg1": agent.Config("CFG1", TEMPLATES["tpl1"], "svc"),
    "cfg2": agent.Config("CFG2", TEMPLATES["tpl2"], "svc")}
NAMESPACES = {
    "ns1": agent.Namespace("NS1"),
    "ns2": agent.Namespace("NS2")}


def _gen_ns_data_fake(namespace):
    return u"{\"DEFAULT\": {\"namespaces\": [[\"" + namespace + u"\", []]]}}"


class TestTemplate(base.TestCase):

    "Test template loading"

    @mock.patch('oslo_config.cfg.open')
    def test_parse_template(self, mock_open):
        "Test loading a template"
        mock.mock_open(mock=mock_open, read_data=NAMESPACE_FILE_CONTENT)
        # Patch the mock_open file to support iteration.
        mock_open.return_value.__iter__ = lambda x: iter(x.readline, '')

        tpl, out_file = agent.Template._parse_template_conf('template')
        self.assertEqual(len(tpl), 2)
        self.assertEqual(out_file, 'etc/svc.conf.sample')


class TestCfgConfig(base.TestCase):

    "Test config handling"

    def test_sanitize(self):
        "test config sanitization"
        conf = cfg.ConfigOpts()
        conf._namespace = cfg._Namespace(conf)
        conf._namespace._normalized = []

        opt_1 = cfg.StrOpt(name='lorem', secret=True)
        opt_2 = cfg.StrOpt(name='ipsum', secret=False)
        conf.register_opts([opt_1, opt_2])

        parsed = {'DEFAULT': {'lorem': ['mysecret'], 'ipsum': ['notsecret']}}
        conf._namespace._normalized.append(parsed)

        agent.Config.sanitize_config(conf)

        self.assertNotIn('mysecret', json.dumps(conf._namespace._normalized))
        self.assertIn(
            'notsecret', json.dumps(conf._namespace._normalized))

        self.assertEqual(conf.lorem, '****')
        self.assertEqual(conf.ipsum, 'notsecret')

    def test_get_info(self):
        "test basic get_info"
        cfg_mock = mock.Mock(spec=agent.Config)
        tpl = mock.Mock()
        tpl.hash = 'lorem'
        cfg_mock.template = tpl
        cfg_mock.path = 'ipsum'
        cfg_mock.data = 'dolor'
        cfg_mock.service = 'svc'

        info = agent.Config.get_info(cfg_mock)

        self.assertIn('template', info)
        self.assertEqual(info['template'], 'lorem')
        self.assertIn('path', info)
        self.assertEqual(info['path'], 'ipsum')
        self.assertIn('data', info)
        self.assertEqual(info['data'], 'dolor')


class TestCfgNamespace(base.TestCase):
    "Test namespace handling"

    @mock.patch('congress.cfg_validator.agent.agent.'
                'validator_generator.generate_ns_data')
    def test_load(self, gen_ns_data_mock):
        "Test load namespace"
        gen_ns_data_mock.return_value = _gen_ns_data_fake('lorem')

        ns_mock = agent.Namespace.load('ipsum')

        self.assertEqual(ns_mock.name, 'ipsum')
        self.assertEqual(json.dumps(ns_mock.data), _gen_ns_data_fake('lorem'))
        self.assertIsNotNone(ns_mock.hash)

        same_data_ns = agent.Namespace.load('other_ipsum')
        self.assertEqual(same_data_ns.hash, ns_mock.hash)

        gen_ns_data_mock.return_value = _gen_ns_data_fake('other_lorem')
        other_data_ns = agent.Namespace.load('ipsum')
        self.assertNotEqual(ns_mock.hash, other_data_ns.hash)

    def test_get_info(self):
        "Test basic info on namespace"
        ns_mock = mock.Mock(spec=agent.Namespace)
        ns_mock.name = 'foo'
        ns_mock.data = 'bar'

        info = agent.Namespace.get_info(ns_mock)

        self.assertIn('name', info)
        self.assertEqual(info['name'], 'foo')
        self.assertIn('data', info)
        self.assertEqual(info['data'], 'bar')


class TestCfgTemplate(base.TestCase):
    """Test the handling of templates"""

    @mock.patch('congress.cfg_validator.agent.agent.'
                'cfg.ConfigOpts._parse_config_files')
    def test_template_conf(self, parse_mock):
        """Test the parsing of template"""
        ns_mock = cfg._Namespace(None)
        ns_mock._normalized.append({
            'DEFAULT':
                {
                    'output_file': ['etc/congress.conf.sample'],
                    'wrap_width': ['79'],
                    'namespace': ['congress', 'oslo.log'],
                }
        })

        parse_mock.return_value = ns_mock

        namespace, out_file = agent.Template._parse_template_conf('somewhere')

        self.assertEqual(out_file, 'etc/congress.conf.sample')
        self.assertIn('congress', namespace)
        self.assertIn('oslo.log', namespace)

    @mock.patch('congress.cfg_validator.agent.agent.Namespace.load')
    @mock.patch('congress.cfg_validator.agent.agent.'
                'Template._parse_template_conf')
    def test_load(self, parse_tpl_conf_mock, load_ns_mock):
        """Test loading a template"""
        parse_tpl_conf_mock.return_value = (['congress_h', 'oslo.log'],
                                            'some/where.sample')

        load_ns_mock.side_effect = [mock.MagicMock(hash='lorem'),
                                    mock.MagicMock(hash='ipsum')]

        tpl = agent.Template.load('path/to/template')

        self.assertEqual(tpl.path, 'path/to/template')
        self.assertEqual(tpl.output_file, 'some/where.sample')
        self.assertEqual(tpl.name, 'where')
        self.assertIsNotNone(tpl.hash)
        self.assertTrue(tpl.namespaces)
        self.assertIn('lorem', [ns.hash for ns in tpl.namespaces])
        self.assertIn('ipsum', [ns.hash for ns in tpl.namespaces])

    def test_get_info(self):
        """Test basic info on template"""
        tpl = mock.Mock(spec=agent.Template)
        tpl.path = 'lorem'
        tpl.output_file = 'ipsum'
        tpl.name = 'dolor'
        ns_mock = mock.Mock()
        ns_mock.hash = 'sit'
        tpl.namespaces = [ns_mock]

        info = agent.Template.get_info(tpl)

        self.assertIn('name', info)
        self.assertEqual(info['name'], 'dolor')
        self.assertIn('path', info)
        self.assertEqual(info['path'], 'lorem')
        self.assertIn('output_file', info)
        self.assertEqual(info['output_file'], 'ipsum')
        self.assertIn('namespaces', info)
        self.assertEqual(info['namespaces'], ['sit'])


def _file_mock(file_spec):
    def _mk_content(spec):
        fval = mock.mock_open(read_data=spec)
        fval.return_value.__iter__ = lambda x: iter(x.readline, '')
        return fval
    file_map = {name: _mk_content(spec)
                for name, spec in six.iteritems(file_spec)}

    def _give_file(name):
        basename = path.basename(name)
        return file_map.get(basename, None)(name)

    return mock.MagicMock(name='open', spec=open, side_effect=_give_file)


def _fake_opt_loader(namespaces):
    def fake_entry_point(namespace):
        if namespace == 'congress':
            return opts_congress.list_opts
        if namespace == 'congress-agent':
            return opts.list_opts
        else:
            return None
    return [(ns, fake_entry_point(ns)) for ns in namespaces]


class TestCfgManager(base.TestCase):
    """Config manager tests"""

    @mock.patch(
        'oslo_config.generator._get_raw_opts_loaders', _fake_opt_loader)
    @mock.patch(
        'oslo_config.cfg.open',
        _file_mock({
            "svc.tpl": NAMESPACE_FILE_CONTENT,
            "svc1.conf": CONF_FILE1_CONTENT,
            "svc2.conf": CONF_FILE2_CONTENT}))
    def test_init(self):
        """Test the creation of the config manager"""
        cfg_manager = agent.ConfigManager(
            'host', {"svc": {"svc1.conf": "svc.tpl", "svc2.conf": "svc.tpl"}})
        self.assertEqual('host', cfg_manager.host)
        self.assertEqual(2, len(cfg_manager.configs))
        self.assertEqual(2, len(cfg_manager.namespaces))
        self.assertEqual(1, len(cfg_manager.templates))
        for conf in six.itervalues(cfg_manager.configs):
            self.assertIsInstance(conf, agent.Config)
        for nspc in six.itervalues(cfg_manager.namespaces):
            self.assertIsInstance(nspc, agent.Namespace)
        for tpl in six.itervalues(cfg_manager.templates):
            self.assertIsInstance(tpl, agent.Template)


def _setup_endpoint():
    with mock.patch(
        'oslo_config.cfg.open',
        _file_mock({"agent.conf": CONF_AGENT_CONTENT})),\
        mock.patch(
            'congress.cfg_validator.agent.agent.ConfigManager',
            autospec=True) as mock_manager,\
        mock.patch(
            'congress.cfg_validator.agent.rpc.ValidatorDriverClient',
            autospec=True) as mock_client:
        conf = cfg.ConfigOpts()
        opts.register_validator_agent_opts(conf)
        conf(args=['--config-file', 'agent.conf'])
        mock_manager.return_value.configs = CONFIGS
        mock_manager.return_value.templates = TEMPLATES
        mock_manager.return_value.namespaces = NAMESPACES
        endpoint = agent.ValidatorAgentEndpoint(conf=conf)
        return endpoint, mock_client, mock_manager


class TestValidatorAgentEndpoint(base.TestCase):
    """Test the endpoint for the agent communications"""

    # pylint: disable=no-self-use

    @mock.patch(
        'oslo_config.cfg.open',
        _file_mock({
            "agent.conf": CONF_AGENT_CONTENT}))
    @mock.patch(
        'congress.cfg_validator.agent.agent.ConfigManager',
        autospec=True)
    @mock.patch(
        'congress.cfg_validator.agent.rpc.ValidatorDriverClient',
        autospec=True)
    def test_publish_template_hashes(self, mock_client, mock_manager):
        "Test a request to publish hashes"
        conf = cfg.ConfigOpts()
        opts.register_validator_agent_opts(conf)
        conf(args=['--config-file', 'agent.conf'])
        templates = {"tpl1": {}, "tpl2": {}}
        mock_manager.return_value.templates = templates
        endpoint = agent.ValidatorAgentEndpoint(conf=conf)
        endpoint.publish_templates_hashes({})
        mock_client.return_value.process_templates_hashes.assert_called_with(
            {}, set(templates), "hhh")

    def test_publish_configs_hashes(self):
        "Test a request to publish hashes"
        endpoint, mock_client, _ = _setup_endpoint()
        endpoint.publish_configs_hashes({})
        mock_client.return_value.process_configs_hashes.assert_called_with(
            {}, set(CONFIGS), "hhh")

    def test_get_config(self):
        "Test reply to an explicit config request"
        endpoint, _, _ = _setup_endpoint()
        ret = endpoint.get_config({}, cfg_hash="cfg1")
        expected = {
            'data': None,
            'path': 'CFG1',
            'service': 'svc',
            'template': None,
            'version': 'vvv'}
        self.assertEqual(expected, ret)
        ret = endpoint.get_config({}, cfg_hash="XXX")
        self.assertEqual(None, ret)

    def test_get_namespace(self):
        "Test reply to an explicit config request"
        endpoint, _, _ = _setup_endpoint()
        ret = endpoint.get_namespace({}, ns_hash="ns1")
        expected = {
            'version': 'vvv',
            'data': None,
            'name': 'NS1'}
        self.assertEqual(expected, ret)
        ret = endpoint.get_namespace({}, ns_hash="XXX")
        self.assertEqual(None, ret)

    def test_get_template(self):
        "Test reply to an explicit config request"
        endpoint, _, _ = _setup_endpoint()
        ret = endpoint.get_template({}, tpl_hash="tpl1")
        expected = {
            'name': 'out',
            'namespaces': [],
            'output_file': 'out',
            'path': 'TPL1',
            'version': 'vvv'}
        self.assertEqual(expected, ret)
        ret = endpoint.get_template({}, tpl_hash="XXX")
        self.assertEqual(None, ret)



#! /usr/bin/env python
#
# Copyright (c) 2014 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
import socket
import sys

from oslo_config import cfg
from oslo_log import log as logging
from oslo_service import service

from congress.common import config
from congress.db import api as db_api
# FIXME It has to initialize distributed_architecture flag basing on the
# config file before the python interpreter imports python file which has
# if-statement for deepsix. Since the default value of the flag is False
# in current implementation, so it will import dse.deepsix as deepsix
# even if you set it to True in congress.conf.
# After changing the default to True, remove following one line.
# This appears in main() too.  Removing either instance breaks something.
config.init(sys.argv[1:])
from congress.common import eventlet_server
from congress import encryption
from congress import harness

LOG = logging.getLogger(__name__)


class ServerWrapper(object):
    """Wraps an eventlet_server with some launching info & capabilities."""

    def __init__(self, server, workers):
        self.server = server
        self.workers = workers

    def launch_with(self, launcher):
        if hasattr(self.server, 'listen'):
            self.server.listen()
        if self.workers > 1:
            # Use multi-process launcher
            launcher.launch_service(self.server, self.workers)
        else:
            # Use single process launcher
            launcher.launch_service(self.server)


def serve(*servers):
    if max([server[1].workers for server in servers]) > 1:
        # TODO(arosen) - need to provide way to communicate with DSE services
        launcher = service.ProcessLauncher(cfg.CONF, restart_method='mutate')
    else:
        launcher = service.ServiceLauncher(cfg.CONF, restart_method='mutate')

    for name, server in servers:
        try:
            server.launch_with(launcher)
        except socket.error:
            LOG.exception(_('Failed to start the %s server'), name)
            raise

    try:
        launcher.wait()
    except KeyboardInterrupt:
        LOG.info("Congress server stopped by interrupt.")


def create_api_server(conf_path, node_id, host, port, workers, policy_engine,
                      datasources):
    congress_api_server = eventlet_server.APIServer(
        conf_path, node_id, host=host, port=port,
        keepalive=cfg.CONF.tcp_keepalive,
        keepidle=cfg.CONF.tcp_keepidle,
        policy_engine=policy_engine,
        api=True,
        datasources=datasources,
        bus_id=cfg.CONF.dse.bus_id)
    # TODO(thinrichs): there's some sort of magic happening for the api
    #   server.  We call eventlet_server, which on start() calls
    #   service.congress_app_factory, which uses harness to create the
    #   API service, which the magic seems to need to do the right thing.
    #   That's why we're not just calling harness.create2 here; instead,
    #   it's buried inside the congress_app_factory.
    return node_id, ServerWrapper(congress_api_server, workers)


def create_nonapi_server(node_id, policy_engine, datasources, workers):
    congress_server = eventlet_server.Server(
        node_id, bus_id=cfg.CONF.dse.bus_id)
    harness.create2(existing_node=congress_server.node, api=False,
                    policy_engine=policy_engine,
                    datasources=datasources)
    return node_id, ServerWrapper(congress_server, workers)


def launch_servers(node_id, api, policy, data):
    servers = []
    if api:
        LOG.info("Starting congress API server on port %d", cfg.CONF.bind_port)
        # API resource runtime encapsulation:
        # event loop -> wsgi server -> webapp -> resource manager
        paste_config = config.find_paste_config()
        config.set_config_defaults()
        servers.append(create_api_server(paste_config,
                                         node_id,
                                         cfg.CONF.bind_host,
                                         cfg.CONF.bind_port,
                                         cfg.CONF.api_workers,
                                         policy_engine=policy,
                                         datasources=data))
    else:
        LOG.info("Starting congress server on node %s", node_id)
        servers.append(create_nonapi_server(node_id, policy, data,
                                            cfg.CONF.api_workers))

    return servers


def main():
    args = sys.argv[1:]

    # TODO(thinrichs): find the right way to do deployment configuration.
    # For some reason we need to config.init(args) in 2 places; here and
    # at the top of the file.  Remove either one, and things break.
    # Note: config.init() will delete the deploy args, so grab them before.
    config.init(args)
    if not cfg.CONF.config_file:
        sys.exit("ERROR: Unable to find configuration file via default "
                 "search paths ~/.congress/, ~/, /etc/congress/, /etc/) and "
                 "the '--config-file' option!")
    if cfg.CONF.replicated_policy_engine and not (
            db_api.is_mysql() or db_api.is_postgres()):
        if db_api.is_sqlite():
            LOG.warning("Deploying replicated policy engine with SQLite "
                        "backend is not officially supported. Unexpected "
                        "behavior may occur. Officially supported backends "
                        "are MySQL and PostgresSQL.")
        else:
            sys.exit("ERROR: replicated_policy_engine option can be used only "
                     "with MySQL or PostgreSQL database backends. Please set "
                     "the connection option in [database] section of "
                     "congress.conf to use a supported backend.")
    config.setup_logging()

    if not (cfg.CONF.api or cfg.CONF.policy_engine or cfg.CONF.datasources):
        # No flags provided, start all services
        cfg.CONF.api = True
        cfg.CONF.policy_engine = True
        cfg.CONF.datasources = True

    # initialize encryption key if datasource services enabled in this instance
    if cfg.CONF.datasources:
        encryption.initialize_key()
        LOG.debug("Initialized encryption key on datasource node")

    # Construct requested deployment
    servers = launch_servers(cfg.CONF.node_id, cfg.CONF.api,
                             cfg.CONF.policy_engine, cfg.CONF.datasources)

    serve(*servers)


if __name__ == '__main__':
    main()



# Copyright (c) 2018 NEC, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
import mock
import sys
import time

sys.modules['monascaclient.client'] = mock.Mock()
sys.modules['monascaclient'] = mock.Mock()

from congress.datasources import monasca_driver
from congress.tests import base
from congress.tests import helper


class TestMonascaDriver(base.TestCase):

    def setUp(self):
        super(TestMonascaDriver, self).setUp()
        self.keystone_client_p = mock.patch(
            "keystoneclient.v3.client.Client")
        self.keystone_client_p.start()
        self.monasca_client_p = mock.patch("monascaclient.client.Client")
        self.monasca_client_p.start()

        args = helper.datasource_openstack_args()
        args['poll_time'] = 0
        args['endpoint'] = 'http://localhost:8070/v2.0'
        self.driver = monasca_driver.MonascaDriver(args=args)

        self.mock_metrics = {"links": [{
            "rel": "self",
            "href": "http://192.168.10.4:8070/v2.0/metrics"}],
            "elements": [{
                "id": "0",
                "name": "mem.used_buffers",
                "dimensions": {
                    "component": "monasca-persister",
                    "service": "monitoring",
                    "hostname": "ceilosca",
                    "url": "http://localhost:8091/metrics"}
            }]}

        self.mock_statistics = {"links": [{
            "rel": "self",
            "href": "http://192.168.10.4:8070/v2.0/metrics/statistics"}],
            "elements": [{
                "id": "2015-11-30T18:00:00Z",
                "name": "mem.used_buffers",
                "dimensions": {},
                "columns": ["timestamp", "avg"],
                "statistics": [
                    ["2015-11-24T00:00:00Z", 56],
                    ["2015-11-24T06:00:00Z", 46],
                    ["2015-11-24T12:00:00Z", 70],
                    ["2015-11-24T18:00:00Z", 60]]
            }]}

    def test_statistics_update_from_datasource(self):
        self.driver._translate_statistics(self.mock_statistics['elements'])
        stats_list = list(self.driver.state[monasca_driver.STATISTICS])
        stats_data_list = list(self.driver.state[monasca_driver.DATA])
        self.assertIsNotNone(stats_list)
        self.assertIsNotNone(stats_data_list)

        expected_stats = [
            ('mem.used_buffers', 'd1fea02438d17fb7446255573bf54d45')]
        self.assertEqual(expected_stats, stats_list)

        expected_stats_data = [
            ('d1fea02438d17fb7446255573bf54d45',
             "['2015-11-24T00:00:00Z', 56]"),
            ('d1fea02438d17fb7446255573bf54d45',
             "['2015-11-24T12:00:00Z', 70]"),
            ('d1fea02438d17fb7446255573bf54d45',
             "['2015-11-24T18:00:00Z', 60]"),
            ('d1fea02438d17fb7446255573bf54d45',
             "['2015-11-24T06:00:00Z', 46]")]
        self.assertEqual(sorted(expected_stats_data), sorted(stats_data_list))

    def test_metrics_update_from_datasource(self):
        with mock.patch.object(self.driver.monasca.metrics, "list") as metrics:
            metrics.return_value = self.mock_metrics['elements']
            self.driver.update_from_datasource()

        expected = {
            'dimensions': set([
                ('e138b5d90a4265c7525f480dd988210b',
                    'component', 'monasca-persister'),
                ('e138b5d90a4265c7525f480dd988210b',
                    'service', 'monitoring'),
                ('e138b5d90a4265c7525f480dd988210b',
                    'hostname', 'ceilosca'),
                ('e138b5d90a4265c7525f480dd988210b',
                    'url', 'http://localhost:8091/metrics')]),
            'metrics': set([
                ('0', 'mem.used_buffers',
                    'e138b5d90a4265c7525f480dd988210b')]),
            'statistics': set([]),
            'statistics.data': set([])
            }
        self.assertEqual(expected, self.driver.state)

    def test_execute(self):
        class MonascaClient(object):
            def __init__(self):
                self.testkey = None

            def getStatistics(self, arg1):
                self.testkey = 'arg1=%s' % arg1

        monasca_client = MonascaClient()
        self.driver.monasca = monasca_client
        api_args = {
            'positional': ['1']
        }
        expected_ans = 'arg1=1'

        self.driver.execute('getStatistics', api_args)

        self.assertEqual(expected_ans, monasca_client.testkey)


class TestMonascaWebhookDriver(base.TestCase):

    def setUp(self):
        super(TestMonascaWebhookDriver, self).setUp()
        self.monasca = monasca_driver.MonascaWebhookDriver('test-monasca')

    @mock.patch.object(monasca_driver.MonascaWebhookDriver, 'publish')
    def test_monasca_webhook_alarm(self, mocked_publish):
        test_alarm = {
            'metrics': [
                {u'dimensions': {u'hostname': u'openstack-13.local.lan',
                                 u'service': u'monitoring'},
                 u'id': None,
                 u'name': u'load.avg_1_min'}],
            'alarm_id': u'3beb4934-053d-4f8f-9704-273bffc2441b',
            'state': u'ALARM',
            'alarm_timestamp': 1531821822,
            'tenant_id': u'3661888238874df098988deab07c599d',
            'old_state': u'UNDETERMINED',
            'alarm_description': u'',
            'message': u'Thresholds were exceeded for the sub-alarms',
            'alarm_definition_id': u'8e5d033f-28cc-459f-91d4-813307e4ca8a',
            'alarm_name': u'alarmPerHost23'}
        self.monasca._webhook_handler(test_alarm)

        expected_rows = set([(u'3beb4934-053d-4f8f-9704-273bffc2441b',
                              u'8e5d033f-28cc-459f-91d4-813307e4ca8a',
                              u'alarmPerHost23',
                              u'',
                              1531821822,
                              u'ALARM',
                              u'UNDETERMINED',
                              u'Thresholds were exceeded for the sub-alarms',
                              u'3661888238874df098988deab07c599d',
                              None,
                              u'load.avg_1_min',
                              u'openstack-13.local.lan',
                              u'monitoring')])
        self.assertEqual(self.monasca.state['alarm_notification'],
                         expected_rows)

    @mock.patch.object(monasca_driver.MonascaWebhookDriver, 'publish')
    def test_webhook_alarm_cleanup(self, mocked_publish):
        self.monasca = monasca_driver.MonascaWebhookDriver(
            'test-monasca',
            args={'hours_to_keep_alarm': 1 / 3600})  # set to 1 sec for test

        test_alarm = {
            'metrics': [
                {u'dimensions': {u'hostname': u'openstack-13.local.lan',
                                 u'service': u'monitoring'},
                 u'id': None,
                 u'name': u'load.avg_1_min'}],
            'alarm_id': u'3beb4934-053d-4f8f-9704-273bffc2441b',
            'state': u'ALARM',
            'alarm_timestamp': 1531821822,
            'tenant_id': u'3661888238874df098988deab07c599d',
            'old_state': u'UNDETERMINED',
            'alarm_description': u'',
            'message': u'Thresholds were exceeded for the sub-alarms',
            'alarm_definition_id': u'8e5d033f-28cc-459f-91d4-813307e4ca8a',
            'alarm_name': u'alarmPerHost23'}

        self.monasca._webhook_handler(test_alarm)

        self.assertEqual(1, len(self.monasca.state['alarm_notification']))
        time.sleep(3)
        self.assertEqual(0, len(self.monasca.state['alarm_notification']))



# Copyright (c) 2014 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import mock
from oslo_db import exception as db_exc

from congress.db import datasources as datasource_db
from congress.dse2 import dse_node
from congress import exception as congressException
from congress.tests.api import base as api_base
from congress.tests import base
from congress.tests import fake_datasource


class TestDataSource(base.SqlTestCase):

    def setUp(self):
        super(TestDataSource, self).setUp()
        config = api_base.setup_config(with_fake_datasource=False, api=False,
                                       policy=False)
        self.dseNode = config['node']
        self.ds_manager = config['ds_manager']

    def _get_datasource_request(self):
        # leave ID out--generated during creation
        return {'name': 'aaron',
                'driver': 'fake_datasource',
                'description': 'hello world!',
                'enabled': True,
                'type': None,
                'config': {'auth_url': 'foo',
                           'username': 'armax',
                           'password': 'password',
                           'tenant_name': 'armax'}}

    def test_add_datasource(self):
        req = self._get_datasource_request()
        result = self.ds_manager.add_datasource(req)
        # test equality of return value except for 'id' field
        del(result['id'])
        self.assertEqual(req, result)
        # check that service actually on dseNode
        services = self.dseNode.get_services()
        self.assertEqual(len(services), 1)
        self.assertEqual(services[0].service_id, req['name'])
        self.assertIsInstance(services[0],
                              fake_datasource.FakeDataSource)
        obj = self.dseNode.invoke_service_rpc(
            req['name'], 'get_status', {'source_id': None, 'params': None})
        self.assertIsNotNone(obj)

    @mock.patch.object(datasource_db, 'add_datasource')
    def test_add_datasource_db_error(self, add_ds):
        add_ds.side_effect = db_exc.DBError('Error in db.')

        req = self._get_datasource_request()
        self.assertRaises(congressException.DatasourceCreationError,
                          self.ds_manager.add_datasource, req)

    @mock.patch.object(dse_node.DseNode, 'register_service')
    def test_add_datasource_synchronizer_error(self, register_ds):
        register_ds.side_effect = Exception('Error in registering service')

        req = self._get_datasource_request()
        self.assertRaises(congressException.DatasourceCreationError,
                          self.ds_manager.add_datasource, req)
        ds = datasource_db.get_datasource_by_name(req['name'])
        self.assertIsNone(ds)

    def test_get_datasource(self):
        req = self._get_datasource_request()
        ds = self.ds_manager.add_datasource(req)
        result = self.dseNode.get_datasource(ds['id'])
        # test equality except for 'id' field
        del(result['id'])
        self.assertEqual(req, result)

    def test_get_datasources(self):
        req = self._get_datasource_request()
        self.ds_manager.add_datasource(req)
        result = self.dseNode.get_datasources()
        self.assertEqual(len(result), 1)
        result = result[0]
        # test equality except for 'id' field
        del(result['id'])
        self.assertEqual(req, result)

    def test_get_datasources2(self):
        req1 = self._get_datasource_request()
        req1['name'] = 'datasource1'
        result1 = self.ds_manager.add_datasource(req1)
        req2 = self._get_datasource_request()
        req2['name'] = 'datasource2'
        result2 = self.ds_manager.add_datasource(req2)
        # check results of add_datasource
        for key, value in req1.items():
            self.assertEqual(value, result1[key])
        for key, value in req2.items():
            self.assertEqual(value, result2[key])
        # check services actually on dseNode
        services = self.dseNode.get_services()
        self.assertEqual(len(services), 2)
        self.assertEqual(set([s.service_id for s in services]),
                         set(['datasource1', 'datasource2']))
        self.assertIsInstance(services[0],
                              fake_datasource.FakeDataSource)
        self.assertIsInstance(services[1],
                              fake_datasource.FakeDataSource)
        # check results of get_datasources
        resultall = self.dseNode.get_datasources()
        self.assertEqual(len(resultall), 2)
        # check equality except for 'id' field
        byname = {x['name']: x for x in resultall}
        for x in byname.values():
            del(x['id'])
        self.assertEqual(byname, {'datasource1': req1, 'datasource2': req2})

    def test_get_datasources_hide_secret(self):
        req = self._get_datasource_request()
        self.ds_manager.add_datasource(req)
        result = self.dseNode.get_datasources(filter_secret=True)
        result = result[0]
        # check equality except that 'config'/'password' is hidden
        req['config']['password'] = "<hidden>"
        del(result['id'])
        self.assertEqual(result, req)

    def test_create_datasource_duplicate_name(self):
        req = self._get_datasource_request()
        self.ds_manager.add_datasource(req)
        self.assertRaises(congressException.DatasourceNameInUse,
                          self.ds_manager.add_datasource, req)

    def test_delete_datasource(self):
        req = self._get_datasource_request()
        result = self.ds_manager.add_datasource(req)
        self.ds_manager.delete_datasource(result)
        # check that service is actually deleted
        services = self.dseNode.get_services()
        self.assertEqual(len(services), 0)
        self.assertRaises(
            congressException.NotFound, self.dseNode.invoke_service_rpc,
            req['name'], 'get_status', {'source_id': None, 'params': None})
        # TODO(thinrichs): test that we've actually removed
        #   the row from the DB

    # TODO(dse2): this test relies on coordination between dseNode and
    #  policy engine.  Much harder in distributed system.  Need to decide
    #  if we want that kind of invariant and if so implement it.
    # def test_delete_datasource_error(self):
    #     req = self._get_datasource_request()
    #     req['driver'] = 'fake_datasource'
    #     req['config'] = {'auth_url': 'foo',
    #                      'username': 'armax',
    #                      'password': 'password',
    #                      'tenant_name': 'armax'}
    #     # let driver generate this for us.
    #     del req['id']
    #     result = self.datasource_mgr.add_datasource(req)
    #     engine = self.dseNode.service_object('engine')
    #     engine.create_policy('alice')
    #     engine.insert('p(x) :- %s:q(x)' % req['name'], 'alice')
    #     self.assertRaises(exception.DanglingReference,
    #                       self.datasource_mgr.delete_datasource,
    #                       result['id'])

    def test_delete_invalid_datasource(self):
        req = self._get_datasource_request()
        req['id'] = 'fake-id'
        self.assertRaises(congressException.DatasourceNotFound,
                          self.ds_manager.delete_datasource, req)

    def test_get_driver_schema(self):
        schema = self.dseNode.get_driver_schema('fake_datasource')
        self.assertEqual(schema,
                         fake_datasource.FakeDataSource.get_schema())
